/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include <shlwapi.h>

#include "ConvertImage.h"
#include "CPU.h"			// AddCPUJob(CPU_CHECK_INTERRUPTS)
#include "D3DRender.h"
#include "Debug.h"
#include "MainWindow.h" // For MainWindow_UpdateScreen()
#include "Dump.h"
#include "ROM.h"			// ROM_Check_Ucode
#include "RDP.h"
#include "RDP_GFX.h"
#include "memory.h"


#include "daedalus_crc.h"
#include "ultra_gbi.h"

// Lkb added mirroring support for cards that don't support it. Thanks Lkb!
/*
Description:
	N64 games often use texture mirroring to display semi-transparent circles.
	Examples:
	- Shadow under Mario and other characters
	- Transition that happens when Mario opens the castle door
	- The cannon in Mario
	- The telescope in Majora's Mask (seen in TR64)

	However, some 3D cards do not support texture mirroring and mirrored textures are displayed incorrectly.
	For the first two cases, this is not a big problem, but it's annoying in the last two.

	This patch emulates mirroring in software when the 3D card doesn't support it (eg. Matrox G400)

	Problems:
	- This patch doesn't redefine any texture coordinate (maybe it's correct, maybe no, I failed to understand this :) )
	- This patch always assumes 2x2 mirroring
	- This patch computes the mirroring at every frame and does not cache the texture
*/

static void RDP_EnableTexturing(DWORD dwTile);
static void RDP_InitRenderState();
static void RDP_GFX_DumpVtxInfo(DWORD dwAddress, DWORD dwV0, DWORD dwN);

static DWORD g_dwNumTrisRendered;
static DWORD g_dwNumDListsCulled;
static DWORD g_dwNumTexturesIgnored;
static DWORD g_dwNumTrisClipped;
static DWORD g_dwNumVertices;


static BOOL g_bDetermineduCode = FALSE;
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                    uCode Config                      //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

// This is the multiplier applied to vertex indices. 
// For Mario 64, it is 10.
// For Starfox, Mariokart etc it is 2.
static DWORD g_dwVertexMult = 10;	

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                      Dumping                         //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
static BOOL g_bDumpNextDisplayList = FALSE;
HANDLE g_hRDPDumpHandle = INVALID_HANDLE_VALUE;

static BOOL g_bDropAllTextures = FALSE;


//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                     GFX State                        //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
static ID3DXContext * s_pD3DX = NULL;
static LPDIRECT3DDEVICE7 s_pD3DDev = NULL;

// Multithread protection for our renderer
static CCritSect g_RendererLock;
static D3DRender * g_Renderer;



typedef struct _SetTImgInfo
{
	DWORD dwFormat;
	DWORD dwSize;
	DWORD dwWidth;
	DWORD dwAddr;
} SetImgInfo;


static DWORD	g_dwSegment[16];
static N64Light  g_N64Lights[8];
static Tile g_Tiles[8];
static SetImgInfo g_TI = { G_IM_FMT_RGBA, G_IM_SIZ_16b, 1, 0 };
static SetImgInfo g_CI = { G_IM_FMT_RGBA, G_IM_SIZ_16b, 1, 0 };
static SetImgInfo g_DI = { G_IM_FMT_RGBA, G_IM_SIZ_16b, 1, 0 };

// The display list PC stack. Before this was an array of 10
// items, but this way we can nest as deeply as necessary. 
static std::vector< u32 > g_dwPCStack;

static BOOL  g_bTextureEnable = FALSE;
static DWORD g_dwGeometryMode = 0;
static DWORD g_dwOtherModeL   = 0;
static DWORD g_dwOtherModeH   = 0;

enum RGDGFX_LOAD_TYPE
{
	LOAD_BLOCK = 0,
	LOAD_TILE
};
static RGDGFX_LOAD_TYPE g_nLastLoad = LOAD_BLOCK;
static DWORD g_dwLoadTilePitch;				// Pitch (as specified at time of LoadTile)
static DWORD g_dwLoadAddress;				// Address (as specified at LoadBlock)

static DWORD g_dwPalAddress = 0;
static DWORD g_dwPalSize = 0;

static DWORD g_dwAmbientLight = 0;

static DWORD g_dwTextureTile = 0;

static struct
{
	DWORD x0, y0, x1, y1, mode;
} g_Scissor;

static BOOL g_bDXTZero = FALSE;					// HACK - need to pass this directly to TH!


static DWORD g_dwFillColor		= 0xFFFFFFFF;




static DWORD g_dwLastPurgeTimeTime = 0;		// Time textures were last purged

static D3DDEVICEDESC7 g_D3DDeviceDesc;




//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                      Strings                         //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

static const char * sc_colcombtypes32[32] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Environment ",
	"1           ", "CombAlp     ",
	"Texel0_Alpha", "Texel1_Alpha",
	"Prim_Alpha  ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Fraction",
	"PrimLODFrac ", "K5          ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ",	"0           "
};
static const char *sc_colcombtypes16[16] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Prim        ", 
	"Shade       ", "Environment ",
	"1           ", "CombAlp     ",
	"Texel0_Alpha", "Texel1_Alpha",
	"Prim_Alp    ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Fraction",
	"PrimLOD_Frac", "0           "
};
static const char *sc_colcombtypes8[8] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Environment ",
	"1           ", "0           ",
};



// Mask down to 0x003FFFFF?
#define RDPSegAddr(seg) ( g_dwSegment[(seg>>24)&0x0F] + (seg&0x00FFFFFF) )


static void RDP_GFX_Nothing(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SPNOOP(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Mtx(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaMoveWordTri(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_MoveMem(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_Vtx_Mario(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Vtx_MarioKart(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Vtx_WaveraceUS(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_Reserved1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_DL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Reserved2(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Reserved3(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Sprite2D(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_Tri2_Mariokart(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Tri2_Goldeneye(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPHalf_Cont(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPHalf_2(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPHalf_1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Line3D(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ClearGeometryMode(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetGeometryMode(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_EndDL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetOtherMode_L(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetOtherMode_H(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Texture(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_MoveWord(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_PopMtx(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_CullDL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Tri1(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_NOOP(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_TexRect(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_TexRectFlip(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPLoadSync(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPPipeSync(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPTileSync(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPFullSync(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetKeyGB(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetKeyR(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetConvert(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetScissor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetPrimDepth(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_RDPSetOtherMode(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_LoadTLut(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetTileSize(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_LoadBlock(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_LoadTile(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetTile(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_FillRect(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetFillColor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetFogColor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetBlendColor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetPrimColor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetEnvColor(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetCombine(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetTImg(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetZImg(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_SetCImg(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_AST_Unk1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_YS_Unk1(DWORD dwCmd0, DWORD dwCmd1);

static void RDP_GFX_ZeldaDL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaCullDL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaEndDL(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaMoveWord(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaTexture(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaSetGeometryMode(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaSetOtherMode_L(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaSetOtherMode_H(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaMoveMem(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaMtx(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaPopMtx(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Vtx_Zelda(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaTri2(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaTri3(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaTri1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Zelda2Unk1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_Zelda2Unk2(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaUnk1(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaUnk2(DWORD dwCmd0, DWORD dwCmd1);
static void RDP_GFX_ZeldaUnk3(DWORD dwCmd0, DWORD dwCmd1);

static RDPInstruction GFXInstruction[256] =
{
	// The First 128 Instructions (0..127) are "DMA" commands (i.e. they
	//  require a DMA transfer to move the required data info position)
	//               Zelda_LoadVtx                            Zelda_CullDL
	RDP_GFX_SPNOOP,	 RDP_GFX_Mtx,     RDP_GFX_ZeldaMoveWordTri, RDP_GFX_MoveMem,
	//                                        ZeldaTri2   ZeldaTri3
	RDP_GFX_Vtx_Mario, RDP_GFX_ZeldaTri1, RDP_GFX_DL,    RDP_GFX_Reserved2,
	RDP_GFX_Reserved3, RDP_GFX_Sprite2D, RDP_GFX_ZeldaUnk3, RDP_GFX_Zelda2Unk2,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//10
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//20
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//30
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//40
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//50
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//60
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//70
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,

	// The next 64 commands (-65 .. -128, or 128..191) are "Immediate" commands,
	//  in that they can be executed immediately with no further memory transfers
//80
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//90
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//a0
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_AST_Unk1,
//b0
	RDP_GFX_Nothing, RDP_GFX_Tri2_Mariokart,    RDP_GFX_RDPHalf_Cont, RDP_GFX_RDPHalf_2,
	RDP_GFX_RDPHalf_1, RDP_GFX_Line3D, RDP_GFX_ClearGeometryMode, RDP_GFX_SetGeometryMode,
	RDP_GFX_EndDL, RDP_GFX_SetOtherMode_L, RDP_GFX_SetOtherMode_H, RDP_GFX_Texture,
	RDP_GFX_MoveWord, RDP_GFX_PopMtx, RDP_GFX_CullDL, RDP_GFX_Tri1,

	// The last 64 commands are "RDP" commands; they are passed through the
	//  RSP and sent to the RDP directly.
//c0
	RDP_GFX_NOOP,    RDP_GFX_Nothing, RDP_GFX_YS_Unk1, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
//d0
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing,
	RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_Nothing, RDP_GFX_ZeldaTexture,
	RDP_GFX_ZeldaPopMtx, RDP_GFX_ZeldaSetGeometryMode, RDP_GFX_ZeldaMtx, RDP_GFX_ZeldaMoveWord,
	RDP_GFX_ZeldaMoveMem, RDP_GFX_ZeldaUnk2, RDP_GFX_ZeldaDL, RDP_GFX_ZeldaEndDL,
//e0
	RDP_GFX_Zelda2Unk1, RDP_GFX_ZeldaUnk1, RDP_GFX_ZeldaSetOtherMode_L, RDP_GFX_ZeldaSetOtherMode_H,
	RDP_GFX_TexRect, RDP_GFX_TexRectFlip, RDP_GFX_RDPLoadSync, RDP_GFX_RDPPipeSync,
	RDP_GFX_RDPTileSync, RDP_GFX_RDPFullSync, RDP_GFX_SetKeyGB, RDP_GFX_SetKeyR,
	RDP_GFX_SetConvert, RDP_GFX_SetScissor, RDP_GFX_SetPrimDepth, RDP_GFX_RDPSetOtherMode,
//f0
	RDP_GFX_LoadTLut, RDP_GFX_Nothing, RDP_GFX_SetTileSize, RDP_GFX_LoadBlock, 
	RDP_GFX_LoadTile, RDP_GFX_SetTile, RDP_GFX_FillRect, RDP_GFX_SetFillColor,
	RDP_GFX_SetFogColor, RDP_GFX_SetBlendColor, RDP_GFX_SetPrimColor, RDP_GFX_SetEnvColor,
	RDP_GFX_SetCombine, RDP_GFX_SetTImg, RDP_GFX_SetZImg, RDP_GFX_SetCImg
};

HRESULT RDP_GFX_Init()
{
	HRESULT hr;
	LPDIRECT3DDEVICE7 pD3DDev;

	g_bDetermineduCode = FALSE;
	g_dwLastPurgeTimeTime = g_dwRDPTime;

	g_bDumpNextDisplayList = FALSE;
	g_hRDPDumpHandle = INVALID_HANDLE_VALUE;


	g_RendererLock.Lock();
		hr = S_OK;

		g_Renderer = new D3DRender();
		if (g_Renderer == NULL)
			return E_OUTOFMEMORY;

		if (SUCCEEDED(hr))
			hr = g_Renderer->Initialize();

		pD3DDev = g_GfxContext.GetD3DDev();
		if (pD3DDev != NULL)
		{
			pD3DDev->GetCaps(&g_D3DDeviceDesc);

		}

	g_RendererLock.Unlock();

	if (FAILED(hr))
		return hr;

	if (!TH_InitTable(8000))
		return E_FAIL;


	return S_OK;
}


HRESULT RDP_GFX_Reset()
{
	g_dwPCStack.clear();

	g_bDetermineduCode = FALSE;

	TH_DropTextures();

	return S_OK;
}


void RDP_GFX_Cleanup()
{
	TH_DestroyTable();

	g_RendererLock.Lock();
	if (g_Renderer != NULL)
	{
		delete g_Renderer;
		g_Renderer = NULL;
	}
	g_RendererLock.Unlock();
}

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                      Logging                         //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////


void RDP_GFX_DumpNextDisplayList()
{
	g_bDumpNextDisplayList = TRUE;
}

void RDP_GFX_DropTextures()
{
	g_bDropAllTextures = TRUE;
}

void RDP_GFX_MakeTexturesBlue()
{
	g_bTHMakeTexturesBlue = TRUE;
	RDP_GFX_DropTextures();
}
void RDP_GFX_MakeTexturesNormal()
{
	g_bTHMakeTexturesBlue = FALSE;
	RDP_GFX_DropTextures();
}

void RDP_GFX_DumpTextures()
{
	DBGConsole_Msg(0, "Texture dumping enabled");
	g_bTHDumpTextures = TRUE;
	RDP_GFX_DropTextures();		// Drop textures so that we can re-create
}
void RDP_GFX_NoDumpTextures()
{
	DBGConsole_Msg(0, "Texture dumping disabled");
	g_bTHDumpTextures = FALSE;
}

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//                   Task Handling                      //
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

#define G_ZELDAVTX  G_MTX			// 1
#define G_ZELDACULLDL  G_MOVEMEM	// 3
// 4 is something like a conditional DL
#define G_ZELDATRI1 0x05		// reserved1
#define G_ZELDATRI2 0x06			// G_DL
#define G_ZELDATRI3 0x07			// reserved

void RDP_GFX_IdentifyuCode(OSTask * pTask)
{
	DBGConsole_Msg(0, "[GReconfiguring RDP to process ucode %d]", g_ROM.ucode);

	// Bugfix from previous v0.03b inifile creation
	if (g_ROM.nScaleFactor <= 0)
		g_ROM.nScaleFactor = 1;

	g_Renderer->SetScaleFactor(g_ROM.nScaleFactor * 1.0f);

	switch (g_ROM.ucode)
	{
	case 0:
		g_dwVertexMult = 10;	
		DBGConsole_Msg(0, "Using uCode 0");
		GFXInstruction[G_MTX] = RDP_GFX_Mtx;
		GFXInstruction[G_MOVEMEM] = RDP_GFX_MoveMem;
		GFXInstruction[G_DL] = RDP_GFX_DL;
		GFXInstruction[G_VTX] = RDP_GFX_Vtx_Mario;
		GFXInstruction[G_TRI2] = RDP_GFX_Tri2_Goldeneye;
		GFXInstruction[7] = RDP_GFX_Reserved2;
		break;
		
	case 1:
		g_dwVertexMult = 2;	
		DBGConsole_Msg(0, "Using uCode 1");
		GFXInstruction[G_MTX] = RDP_GFX_Mtx;
		GFXInstruction[G_MOVEMEM] = RDP_GFX_MoveMem;
		GFXInstruction[G_DL] = RDP_GFX_DL;
		GFXInstruction[G_VTX] = RDP_GFX_Vtx_MarioKart;
		GFXInstruction[G_TRI2] = RDP_GFX_Tri2_Mariokart;
		GFXInstruction[7] = RDP_GFX_Reserved2;
		break;

	case 4:
		g_dwVertexMult = 5;	
		DBGConsole_Msg(0, "Using uCode 4");
		GFXInstruction[G_MTX] = RDP_GFX_Mtx;
		GFXInstruction[G_MOVEMEM] = RDP_GFX_MoveMem;
		GFXInstruction[G_DL] = RDP_GFX_DL;
		GFXInstruction[G_VTX] = RDP_GFX_Vtx_WaveraceUS;
		GFXInstruction[G_TRI2] = RDP_GFX_Tri2_Mariokart;		// Not used?
		GFXInstruction[7] = RDP_GFX_Reserved2;
		break;

	case 5:
		g_dwVertexMult = 2;	
		DBGConsole_Msg(0, "Using uCode 5");
		GFXInstruction[G_ZELDAVTX] = RDP_GFX_Vtx_Zelda;
		GFXInstruction[G_ZELDACULLDL] = RDP_GFX_ZeldaCullDL;
		GFXInstruction[G_ZELDATRI2] = RDP_GFX_ZeldaTri2;
		GFXInstruction[7] = RDP_GFX_ZeldaTri3;
		GFXInstruction[G_VTX] = RDP_GFX_Vtx_Mario;
		GFXInstruction[G_TRI2] = RDP_GFX_Tri2_Goldeneye;
		break;


	default:
		g_dwVertexMult = 10;	
		DBGConsole_Msg(0, "Using default uCode");
		GFXInstruction[G_MTX] = RDP_GFX_Mtx;
		GFXInstruction[G_MOVEMEM] = RDP_GFX_MoveMem;
		GFXInstruction[G_DL] = RDP_GFX_DL;
		GFXInstruction[G_VTX] = RDP_GFX_Vtx_Mario;
		GFXInstruction[G_TRI2] = RDP_GFX_Tri2_Goldeneye;

	}
}




void RDP_GFX_ExecuteTask(OSTask * pTask)
{
	DWORD dwPC;
	DWORD dwCmd0;
	DWORD dwCmd1;
	HRESULT hr;		
	DWORD dwWidth;
	DWORD dwHeight;

	if (g_bDetermineduCode == FALSE)
	{
		RDP_GFX_IdentifyuCode(pTask);
		g_bDetermineduCode = TRUE;
	}

	// Initialise stack
	g_dwPCStack.clear();
	g_dwPCStack.push_back((u32)pTask->t.data_ptr);


	// Check if we need to purge
	if (g_dwRDPTime - g_dwLastPurgeTimeTime > 5000)
	{
		TH_PurgeOldTextures();
		g_dwLastPurgeTimeTime = g_dwRDPTime;
	}

	g_dwNumTexturesIgnored = 0;
	g_dwNumDListsCulled = 0;
	g_dwNumTrisRendered = 0;
	g_dwNumTrisClipped = 0;
	g_dwNumVertices = 0;

	if (g_bDumpNextDisplayList)
	{
		static DWORD dwCount = 0;

		TCHAR szFilePath[MAX_PATH+1];
		TCHAR szFileName[MAX_PATH+1];
		TCHAR szDumpDir[MAX_PATH+1];

		wsprintf(szDumpDir, "%s\\%s", TEXT("dl_dumps"), g_ROM.szGameName);
		Dump_GetDumpDirectory(szFilePath, szDumpDir);

		wsprintf(szFileName, "dl%04d.txt", dwCount++);

		PathAppend(szFilePath, szFileName);

		g_hRDPDumpHandle = Debug_CreateDumpFile(szFilePath);
		if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
			DBGConsole_Msg(0, "RDP: Dumping Display List as %s", szFilePath);
		else
			DBGConsole_Msg(0, "RDP: Couldn't create dumpfile %s", szFilePath);

		// Clear flag as we're done
		g_bDumpNextDisplayList = FALSE;
	}

	if (g_bDropAllTextures)
	{
		TH_DropTextures();
		g_bDropAllTextures = FALSE;
	}
		

	DL_PF("DP: Firing up RDP!");

	// Lock the graphics context here.
	g_GfxContext.Lock();
	g_RendererLock.Lock();
	if (g_bRDPEnableGfx && g_GfxContext.Ready() && g_Renderer != NULL)
	{
		s_pD3DX = g_GfxContext.GetD3DX();
		s_pD3DDev = g_GfxContext.GetD3DDev();
		if (s_pD3DX)
			s_pD3DX->AddRef();
		if (s_pD3DDev)
			s_pD3DDev->AddRef();

		if (s_pD3DX != NULL && s_pD3DDev != NULL)
		{
			s_pD3DX->AddRef();

			// ZBuffer clearing is caught elsewhere
			//s_pD3DX->Clear(D3DCLEAR_ZBUFFER);
			//s_pD3DX->Clear(D3DCLEAR_TARGET);
			
			hr = s_pD3DX->GetBufferSize(&dwWidth, &dwHeight);
			if (SUCCEEDED(hr))
				g_Renderer->SetScreenMult((float)dwWidth / (float)g_dwViWidth,
										  (float)dwHeight / (float)g_dwViHeight);
			else
				g_Renderer->SetScreenMult(1.0f, 1.0f);
			
			g_Renderer->Reset();

			s_pD3DDev->BeginScene();

			g_Renderer->SetViewport(0, 0, g_dwViWidth, g_dwViHeight);

			// Process the entire display list in one go
			while (!g_bRDPHalted)
			{
				// Current PC is the last value on the stack
				dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

				if (dwPC > g_dwRamSize)
				{
					DBGConsole_Msg(0, "Display list PC is out of range: 0x%08x", dwPC);
					break;
				}

				dwCmd0 = g_pu32RamBase[(dwPC>>2)+0];
				dwCmd1 = g_pu32RamBase[(dwPC>>2)+1];

				g_dwPCStack[g_dwPCStack.size() - 1] += 8;

				try 
				{
					GFXInstruction[dwCmd0>>24](dwCmd0, dwCmd1);
				}
				catch (...)
				{
					DBGConsole_Msg(0, "Exception processing 0x%08x 0x%08x", dwCmd0, dwCmd1);
					if (!g_bTrapExceptions)
						throw;
					break;
				}
			}

			s_pD3DDev->EndScene();
		}

		// One of the other conditions may have failed above
		// so this may still be a valid pointer
		SAFE_RELEASE(s_pD3DDev);
		SAFE_RELEASE(s_pD3DX);
	}

	/*ID3DXContext * pD3DX = g_GfxContext.GetD3DX();
	if (pD3DX)
	{
		LPDIRECTDRAWSURFACE7 pSurf = pD3DX->GetBackBuffer(0);
		if (pSurf)
		{
			HRESULT hr;
			// Hack to avoid splattering memory on first DL
			if (Memory_VI_GetRegister(VI_ORIGIN_REG) > 0x280)
			{
				DDSURFACEDESC2 ddsd;
				ZeroMemory( &ddsd, sizeof(DDSURFACEDESC2) );
				ddsd.dwSize = sizeof(ddsd);
				hr = pSurf->Lock(NULL, &ddsd, DDLOCK_NOSYSLOCK|DDLOCK_WAIT , NULL);

				if (SUCCEEDED(hr))
				{
					LONG y, x;
					u16 * pDst = &g_pu16RamBase[Memory_VI_GetRegister(VI_ORIGIN_REG)/2];

					for (y = 0; y < g_dwViHeight; y++)
					{
						u8 *pSrc = (u8*)ddsd.lpSurface + ((y * dwHeight) / g_dwViHeight) * ddsd.lPitch;
						for (x = 0; x < g_dwViWidth; x++)
						{
							//*p++ = 0x0001;
							pDst[x^0x2] = *(u16*)(pSrc + ((x * dwWidth) / g_dwViWidth));
						}
						pDst = (u16*)((u8*)pDst + x/2);
						//pSrc = (u16*)((u8*)pSrc + ddsd.lPitch);

					}

					pSurf->Unlock(NULL);
				}
			}

			//pSurf->Release();

		}
	}*/




	g_RendererLock.Unlock();
	g_GfxContext.Unlock();


	// Do this regardless!
	Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_DP);
	AddCPUJob(CPU_CHECK_INTERRUPTS);

	if (g_dwNumTrisRendered + g_dwNumTrisClipped > 0)
	{
		DBGConsole_Stats(2, "%dv %dt (%dclip)",
			g_dwNumVertices, g_dwNumTrisRendered + g_dwNumTrisClipped, g_dwNumTrisClipped);
	}
	DBGConsole_Stats(8, "%d tex load ignored", g_dwNumTexturesIgnored);
	DBGConsole_Stats(9, "%d dlists culled", g_dwNumDListsCulled);

	// Have image displayed
//	if (g_dwNumTrisRendered > 0)
		//MainWindow_UpdateScreen();

	DPF(DEBUG_MEMORY_PI, "******************* FRAME **************");
	// Call RSP BREAK here.
	DL_PF("EndDisplayList: Finished - Calling RSP::BREAK.");


	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		Debug_CloseDumpFile(g_hRDPDumpHandle);

		g_hRDPDumpHandle = INVALID_HANDLE_VALUE;
	}


//	DBGConsole_Msg(0, "Frame");

	g_dwNumFrames++;

}

void RDP_MirrorEmulator_DrawLine(DrawInfo& destInfo, DrawInfo& srcInfo, LPDWORD pSource, LPDWORD pDest, DWORD nWidth, BOOL bFlipLeftRight)
{
	if(!bFlipLeftRight)
	{
		memcpy(pDest, pSource, nWidth * 4);
	}
	else
	{
		LPDWORD pMaxDest = pDest + nWidth;
		pSource += nWidth - 1;
		for(; pDest < pMaxDest; pDest++, pSource--)
		{
			*pDest = *pSource;
		}
	}
}

void RDP_MirrorEmulator_Draw(DrawInfo& destInfo, DrawInfo& srcInfo, DWORD nDestX, DWORD nDestY, BOOL bFlipLeftRight, BOOL bFlipUpDown)
{
	LPBYTE pDest = (LPBYTE)((DWORD)destInfo.lpSurface + (destInfo.lPitch * nDestY) + (4 * nDestX));
	LPBYTE pMaxDest = pDest + (destInfo.lPitch * srcInfo.dwHeight);
	LPBYTE pSource = (LPBYTE)(srcInfo.lpSurface);
	if(!bFlipUpDown)
	{
		for(; pDest < pMaxDest; pDest += destInfo.lPitch, pSource += srcInfo.lPitch)
		{
			RDP_MirrorEmulator_DrawLine(destInfo, srcInfo, (LPDWORD)pSource, (LPDWORD)pDest, srcInfo.dwWidth, bFlipLeftRight);
		}
	}
	else
	{
		pSource += (srcInfo.lPitch * (srcInfo.dwHeight - 1));
		for(; pDest < pMaxDest; pDest += destInfo.lPitch, pSource -= srcInfo.lPitch)
		{
			RDP_MirrorEmulator_DrawLine(destInfo, srcInfo, (LPDWORD)pSource, (LPDWORD)pDest, srcInfo.dwWidth, bFlipLeftRight);
		}
	}
}

SurfaceHandler* RDP_GetSurfaceHandler(DWORD dwTile, TextureEntry *pEntry)
{
	if((g_Tiles[dwTile].bMirrorS) && (g_Tiles[dwTile].bMirrorT) && !(g_D3DDeviceDesc.dpcTriCaps.dwTextureAddressCaps & D3DPTADDRESSCAPS_MIRROR))
	{
		if(pEntry->pMirroredSurface)
		{
			return pEntry->pMirroredSurface;
		}
		else
		{
			SurfaceHandler* pSurfaceHandler = NULL;

			// FIXME: Compute the correct values. 2/2 seems to always work correctly in Mario64
			DWORD nXTimes = 2;
			DWORD nYTimes = 2;
			
			DrawInfo srcInfo;		
			pEntry->pSurface->StartUpdate(&srcInfo);
			DWORD nWidth = srcInfo.dwWidth;
			DWORD nHeight = srcInfo.dwHeight;

			pSurfaceHandler = new SurfaceHandler(nWidth * nXTimes, nHeight * nYTimes);
			DrawInfo destInfo;
			pSurfaceHandler->StartUpdate(&destInfo);
			
			
			for(DWORD nY = 0; nY < nYTimes; nY++)
			{
				for(DWORD nX = 0; nX < nXTimes; nX++)
				{
					RDP_MirrorEmulator_Draw(destInfo, srcInfo, nWidth * nX, nHeight * nY, nX & 0x1, nY & 0x1);
				}
			}

			pSurfaceHandler->EndUpdate(&destInfo);
			pEntry->pSurface->EndUpdate(&srcInfo);

			pEntry->pMirroredSurface = pSurfaceHandler;

			return pSurfaceHandler;
		}
	}
	return NULL;
}

void RDP_EnableTexturing(DWORD dwTile)
{
	DWORD dwTileWidth;
	DWORD dwTileHeight;
	DWORD dwTLutFmt;
	DWORD dwPitch;
	DWORD dwPalOffset;
	TextureEntry *pEntry = NULL;

	dwTLutFmt = g_dwOtherModeH& (0x3<<G_MDSFT_TEXTLUT);

	// Now initialise surfaces
	dwTileWidth = (g_Tiles[dwTile].nLRS - g_Tiles[dwTile].nULS);
	dwTileHeight= (g_Tiles[dwTile].nLRT - g_Tiles[dwTile].nULT);

	DL_PF("Enabling Texturing for Tile: %d", dwTile);

	if (dwTileWidth == 0 || dwTileHeight == 0)
	{
		DL_PF("Ignoring zero-sized tile load");
		//DBGConsole_Msg(0, "Ignoring zero-sized tile load");
		return;
	}



	dwPalOffset = 0;

	// Only needs doing for CI, but never mind
	switch (g_Tiles[dwTile].dwSize)
	{
	case G_IM_SIZ_4b: dwPalOffset = 16  * 2 * g_Tiles[dwTile].dwPalette; break;
	case G_IM_SIZ_8b: dwPalOffset = 256 * 2 * g_Tiles[dwTile].dwPalette; break;
	}	

	
	if (g_nLastLoad == LOAD_BLOCK)
	{
		// HACK: In Mario, the texture for the Head bit is offset for some reason
		g_Tiles[dwTile].nLRS -= g_Tiles[dwTile].nULS;
		g_Tiles[dwTile].nLRT -= g_Tiles[dwTile].nULT;			
		g_Tiles[dwTile].nULS = 0;
		g_Tiles[dwTile].nULT = 0;			

		// It was a block load - the pitch is determined by the tile size
		dwPitch = g_Tiles[dwTile].dwPitch;
	}
	else
	{
		// It was a tile - the pitch is set when the tile is loaded
		g_bDXTZero = 0;
		dwPitch = g_dwLoadTilePitch;
	}

	DL_PF("Performing tmap load:");
	DL_PF("     (%d,%d -> %d,%d)",
						g_Tiles[dwTile].nULS, g_Tiles[dwTile].nULT, 
						g_Tiles[dwTile].nLRS, g_Tiles[dwTile].nLRT);
	DL_PF("     Pitch: %d, Addr: 0x%08x", dwPitch, g_dwLoadAddress);



	GetTextureInfo gti;


	gti.dwAddress = g_dwLoadAddress;
	gti.dwFormat = g_Tiles[dwTile].dwFormat;
	gti.dwSize = g_Tiles[dwTile].dwSize;
	gti.nLeft = g_Tiles[dwTile].nULS;
	gti.nTop = g_Tiles[dwTile].nULT;
	gti.dwWidth = dwTileWidth;
	gti.dwHeight = dwTileHeight;
	gti.dwPitch = dwPitch;
	gti.dwPalAddress = g_dwPalAddress + dwPalOffset;
	gti.dwPalSize = g_dwPalSize;
	gti.dwTLutFmt = dwTLutFmt;
	gti.bSwapped = g_bDXTZero;

	// Hack - Extreme-G specifies RGBA/8 textures, but they're really CI8
	if (gti.dwFormat == G_IM_FMT_RGBA &&
		(gti.dwSize == G_IM_SIZ_8b || gti.dwSize == G_IM_SIZ_4b))
		gti.dwFormat = G_IM_FMT_CI;

	if (gti.dwFormat == G_IM_FMT_CI)
	{
		if (gti.dwTLutFmt == G_TT_NONE)
			gti.dwTLutFmt = G_TT_RGBA16;		// Force RGBA

	}

	pEntry = TH_GetTexture(&gti);
	if (pEntry != NULL && pEntry->pSurface != NULL)
	{	
		//Lkb mirror support
		SurfaceHandler* pSurfaceHandler = NULL;
		
		pSurfaceHandler = RDP_GetSurfaceHandler(dwTile, pEntry);	

		g_Renderer->SetTexture( (pSurfaceHandler != NULL)? pSurfaceHandler->GetSurface() : pEntry->pSurface->GetSurface(),
								pEntry->dwWidth,
								pEntry->dwHeight);

		delete pSurfaceHandler;

		// Initialise the clamping state...what if clamp AND mirror?
		if (g_Tiles[dwTile].bMirrorS)		g_Renderer->SetAddressU( 0, D3DTADDRESS_MIRROR );
		else if (g_Tiles[dwTile].bClampS)	g_Renderer->SetAddressU( 0, D3DTADDRESS_CLAMP );
		else								g_Renderer->SetAddressU( 0, D3DTADDRESS_WRAP );
		
		if (g_Tiles[dwTile].bMirrorT)		g_Renderer->SetAddressV( 0, D3DTADDRESS_MIRROR );
		else if (g_Tiles[dwTile].bClampT)	g_Renderer->SetAddressV( 0, D3DTADDRESS_CLAMP );
		else								g_Renderer->SetAddressV( 0, D3DTADDRESS_WRAP );
			
	}
	else
	{
		g_Renderer->SetTexture( NULL, 64, 64 );
	}
}





void RDP_MoveMemLight(DWORD dwLight, DWORD dwAddress)
{
	s8 * pcBase = g_ps8RamBase + dwAddress;
	DWORD * pdwBase = (DWORD *)pcBase;

	g_N64Lights[dwLight].dwRGBA     = pdwBase[0];
	g_N64Lights[dwLight].dwRGBACopy = pdwBase[1];
	g_N64Lights[dwLight].x			= (float)pcBase[8 ^ 0x3];
	g_N64Lights[dwLight].y			= (float)pcBase[9 ^ 0x3];
	g_N64Lights[dwLight].z			= (float)pcBase[10 ^ 0x3];
					
	DL_PF("       RGBA: 0x%08x, RGBACopy: 0x%08x, x: %f, y: %f, z: %f", 
		g_N64Lights[dwLight].dwRGBA,
		g_N64Lights[dwLight].dwRGBACopy,
		g_N64Lights[dwLight].x,
		g_N64Lights[dwLight].y,
		g_N64Lights[dwLight].z);

	if (dwLight == g_dwAmbientLight)
	{
		DL_PF("      (Ambient Light)");

		DWORD dwCol = RGB_MAKE( (g_N64Lights[dwLight].dwRGBA >> 24)&0xFF,
					  (g_N64Lights[dwLight].dwRGBA >> 16)&0xFF,
					  (g_N64Lights[dwLight].dwRGBA >>  8)&0xFF );

		g_Renderer->SetAmbientLight( dwCol );
	}
	else
	{
		
		DL_PF("      (Normal Light)");

		g_Renderer->SetLightCol(dwLight, g_N64Lights[dwLight].dwRGBA);
		if (pdwBase[2] == 0)	// Direction is 0!
		{
			DL_PF("      Light is invalid");
		}
		else
		{
			g_Renderer->SetLightDirection(dwLight, 
										  -g_N64Lights[dwLight].x,
										  -g_N64Lights[dwLight].y,
										  -g_N64Lights[dwLight].z);
		}
	}
}

void RDP_MoveMemViewport(DWORD dwAddress)
{
	s16 scale[4];
	s16 trans[4];

	// dwAddress is offset into RD_RAM of 8 x 16bits of data...
	scale[0] = *(s16 *)(g_pu8RamBase + ((dwAddress+(0*2))^0x2));
	scale[1] = *(s16 *)(g_pu8RamBase + ((dwAddress+(1*2))^0x2));
	scale[2] = *(s16 *)(g_pu8RamBase + ((dwAddress+(2*2))^0x2));
	scale[3] = *(s16 *)(g_pu8RamBase + ((dwAddress+(3*2))^0x2));

	trans[0] = *(s16 *)(g_pu8RamBase + ((dwAddress+(4*2))^0x2));
	trans[1] = *(s16 *)(g_pu8RamBase + ((dwAddress+(5*2))^0x2));
	trans[2] = *(s16 *)(g_pu8RamBase + ((dwAddress+(6*2))^0x2));
	trans[3] = *(s16 *)(g_pu8RamBase + ((dwAddress+(7*2))^0x2));


	LONG nCenterX = trans[0]/4;
	LONG nCenterY = trans[1]/4;
	LONG nWidth   = scale[0]/4;
	LONG nHeight  = scale[1]/4;


	LONG nLeft = nCenterX - nWidth;
	LONG nTop  = nCenterY - nHeight;
	LONG nRight= nCenterX + nWidth;
	LONG nBottom= nCenterY + nHeight;

	// Stupid hack- D3D insists that the vp coords are positive, so
	// we have to truncate them to 0. This happens a lot, as things
	// seem to specify the scale as the screen w/2 h/2
	if (nLeft < 0)
	{
		nRight += nLeft;
		nLeft = 0;
	}
	if (nTop < 0)
	{
		nBottom += nTop;
		nTop = 0;
	}
	if (nRight > g_dwViWidth)
		nRight = g_dwViWidth;
	if (nBottom > g_dwViHeight)
		nBottom = g_dwViHeight;

	g_Renderer->SetViewport(nLeft, nTop, nRight, nBottom);


	DL_PF("        Scale: %d %d %d %d = %d,%d", scale[0], scale[1], scale[2], scale[3], nWidth, nHeight);
	DL_PF("        Trans: %d %d %d %d = %d,%d", trans[0], trans[1], trans[2], trans[3], nCenterX, nCenterY);
	//nLeft = nCenterX - 10;
	//nTop  = nCenterY - 10;
	//nRight= nCenterX + 10;
	//nBottom= nCenterY + 10;
	//g_Renderer->FillRect(nLeft, nTop, nRight, nBottom, RGBA_MAKE(255,255,255,80));
}

#define RDP_NOIMPL_WARN(op) { DBGConsole_Msg(0, op); }
//#define RDP_NOIMPL_WARN 1 ? (void)0 : (void)
inline void RDP_NOIMPL(LPCTSTR op, DWORD dwCmd0, DWORD dwCmd1) 
{
	DBGConsole_Msg(0, "Not Implemented: %s 0x%08x 0x%08x", op, dwCmd0, dwCmd1);
	DL_PF("~~RDP: 0x%08x 0x%08x Not Implemented: %s", dwCmd0, dwCmd1, op);

	/*RDPHalt();*/
}

//Nintro64 uses Sprite2d 
void RDP_GFX_Nothing(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("~~RDP Command Does not exist... 0x%08x 0x%08x", dwCmd0, dwCmd1);

	// Terminate!
	{
		DBGConsole_Msg(0, "Warning, DL cut short with unknown command: 0x%08x 0x%08x",
			dwCmd0, dwCmd1);
		RDPHalt();
		g_dwPCStack.clear();
	}

}

void RDP_GFX_Zelda2Unk1(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("Zelda2Unk1: 0x%08x 0x%08x", dwCmd0, dwCmd1);
}
void RDP_GFX_Zelda2Unk2(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("Zelda2Unk2: 0x%08x 0x%08x", dwCmd0, dwCmd1);
}

void RDP_GFX_ZeldaUnk1(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaUnk1: 0x%08x 0x%08x", dwCmd0, dwCmd1);
}
void RDP_GFX_ZeldaUnk2(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaUnk2: 0x%08x 0x%08x", dwCmd0, dwCmd1);
}
void RDP_GFX_ZeldaUnk3(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaUnk3: 0x%08x 0x%08x", dwCmd0, dwCmd1);
}


void RDP_GFX_SPNOOP(DWORD dwCmd0, DWORD dwCmd1)		{  }


/*
Zelda: Reserved0(0x02140016 0x03e001e0)
Zelda: Reserved0(0x02140014 0x08400300)
Zelda: Reserved0(0x02140010 0x09000500)
Zelda: Reserved0(0x02140002 0x092007c0)
Zelda: Reserved0(0x0214001c 0x086001c0)
Zelda: Reserved0(0x0214002c 0x01000120)
Zelda: Reserved0(0x0214001c 0x006001c0)
Zelda: Reserved0(0x02140026 0x02600120)
Zelda: Reserved0(0x02140028 0x040000c0)
*/

/*
typedef struct {
	short		ob[3];	/* x, y, z * /
	unsigned short	flag;
	short		tc[2];	/* texture coord * /
	unsigned char	cn[4];	/* color & alpha * /
} Vtx_t;

/*
 * Vertex (set up for use with normals)
 * /
typedef struct {
	short		ob[3];	/* x, y, z * /
	unsigned short	flag;
	short		tc[2];	/* texture coord * /
	signed char	n[3];	/* normal * /
	unsigned char   a;      /* alpha  * /
} Vtx_tn;*/

void RDP_GFX_ZeldaMoveWordTri(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwOffset = (dwCmd0 >> 16) & 0xFF;
	DWORD dwVert   = ((dwCmd0      ) & 0xFFFF) / 2;
	DWORD dwValue  = dwCmd1;

	// Data for other commands?
	DL_PF("Zelda: MoveWordTri(0x%08x 0x%08x)", dwCmd0, dwCmd1);

	switch (dwOffset)
	{
	case 0x00:		// Pos x/y
	case 0x04:		// Pos z/flag
	case 0x08:		// Texture coords
	case 0x0c:		// Color/Alpha/Normal
		DBGConsole_Msg(0, "Zelda MoveWordTri - Setting vert data 0x%02x, 0x%08x", dwOffset, dwValue);
		DL_PF("      Setting vert data 0x%02x, 0x%08x", dwOffset, dwValue);
		break;
	default:
		DL_PF("      Setting unk value: 0x%02x, 0x%08x", dwOffset, dwValue);
		break;
	}
}

// YoshiStory does this - 0xc2
void RDP_GFX_YS_Unk1(DWORD dwCmd0, DWORD dwCmd1)
{
	// 
	static BOOL bWarned = FALSE;

	DL_PF("RDP_GFX_YS_Unk1(0x%08x 0x%08x)", dwCmd0, dwCmd1);
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: RDP_GFX_YS_Unk1", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}

}
// AllStarTennis does this - 0xaf...
void RDP_GFX_AST_Unk1(DWORD dwCmd0, DWORD dwCmd1)
{
	// 
	static BOOL bWarned = FALSE;

	DL_PF("RDP_GFX_AST_Unk1(0x%08x 0x%08x)", dwCmd0, dwCmd1);
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: RDP_GFX_AST_Unk1", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}

}

void RDP_GFX_Reserved1(DWORD dwCmd0, DWORD dwCmd1)
{
	// 
	static BOOL bWarned = FALSE;

	DL_PF("Reserved1(0x%08x 0x%08x)", dwCmd0, dwCmd1);
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: Reserved1", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}

}
void RDP_GFX_Reserved2(DWORD dwCmd0, DWORD dwCmd1)
{
	// BB2k
	static BOOL bWarned = FALSE;

	DL_PF("Reserved2(0x%08x 0x%08x)", dwCmd0, dwCmd1);
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: Reserved2", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}

}
void RDP_GFX_Reserved3(DWORD dwCmd0, DWORD dwCmd1)
{		
	// Spiderman
	static BOOL bWarned = FALSE;

	DL_PF("Reserved3(0x%08x 0x%08x)", dwCmd0, dwCmd1);
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: Reserved3", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}
}

void RDP_GFX_Sprite2D(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAddress = RDPSegAddr(dwCmd1);

	DL_PF("Sprite2D: 0x%08x 0x%08x  Address:0x%08x", dwCmd0, dwCmd1, dwAddress);

}

void RDP_GFX_CullDL(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("CullDL: 0x%08x, 0x%08x", dwCmd0, dwCmd1);
}

void RDP_GFX_ZeldaCullDL(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD i;
	DWORD dwVFirst = ((dwCmd0) & 0x1f) / 2;
	DWORD dwVLast  = ((dwCmd1) & 0x1f) / 2;

	DL_PF("ZeldaCullDL: 0x%08x, 0x%08x. Culling using verts %d to %d",
		dwCmd0, dwCmd1, dwVFirst, dwVLast);


	for (i = dwVFirst; i <= dwVLast; i++)
	{
		if (g_Renderer->m_dwVecFlags[i] == 0)
		{
			DL_PF("   Vertex %d is visible, returning", i);
			return;
		}
	}

	g_dwNumDListsCulled++;

	DL_PF("      No vertices were visible, culling");

	DL_PF("      Returning from DisplayList");
	DL_PF("#############################################");
	DL_PF("");
	DL_PF("");

	if (g_dwPCStack.size() == 0)
	{
		DL_PF("EndDisplayList: Too many EndDL calls - ignoring");
		RDPHalt();			// Stop, damn it!
		return;
	}

	// Check if this is the lasy display list in the sequence
	if(g_dwPCStack.size() == 1)
	{
		RDPHalt();
		g_dwPCStack.clear();
		return;
	}

	// If we're here, then we musn't have finished with the display list yet
	g_dwPCStack.pop_back();

}


void RDP_GFX_NOOP(DWORD dwCmd0, DWORD dwCmd1)
{
}
void RDP_GFX_SetKeyGB(DWORD dwCmd0, DWORD dwCmd1)
{
	static BOOL bWarned = FALSE;

	DL_PF("SetKeyGB (not implemented)");
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: SetKeyGB", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}

}
void RDP_GFX_SetKeyR(DWORD dwCmd0, DWORD dwCmd1)
{
	static BOOL bWarned = FALSE;

	DL_PF("SetKeyR (not implemented)");
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: SetKeyR", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}
}

void RDP_GFX_SetConvert(DWORD dwCmd0, DWORD dwCmd1)
{
	
	// Spiderman
	static BOOL bWarned = FALSE;

	DL_PF("SetConvert (not implemented)");
	if (!bWarned)
	{
		RDP_NOIMPL("RDP: SetConvert", dwCmd0, dwCmd1);
		bWarned = TRUE;
	}


}
void RDP_GFX_SetPrimDepth(DWORD dwCmd0, DWORD dwCmd1)
{
	static BOOL bWarned = FALSE;

	DWORD dwZ  = (dwCmd1 >> 16) & 0xFFFF;
	DWORD dwDZ = (dwCmd1      ) & 0xFFFF;

	if (!bWarned)
	{
		RDP_NOIMPL("RDP: SetPrimDepth", dwCmd0, dwCmd1);

		bWarned = TRUE;
	}
	
	DL_PF("SetPrimDepth: 0x%08x 0x%08x - z: 0x%04x dz: 0x%04x",
		dwCmd0, dwCmd1, dwZ, dwDZ);
		
}

//Zelda used:
void RDP_GFX_RDPSetOtherMode(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaRDP: RDPSetOtherMode", dwCmd0, dwCmd1);
}




/* Goldeneye Palettes:
// Works when * 1
SetTextureImage: Image: 0x001af2f0 Fmt: CI/16 Width: 1
SetTile: 0xf5500000 0x07000000 - Tile:7  Fmt: CI/16bpp Line:0 (Pitch 0) TMem:0x0000 Palette:0
         S: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
         T: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
LoadBlock: 0xf3000000 0x072bb000 pad(0x00)
           Tile:7 (0,0) 700 pixels DXT:0x0000 = 1 QWs => 4 pixels/line
          Offset: 0x001af2f0

PipeSync

SetTile: 0xf5000300 0x07000000 - Tile:7  Fmt: RGBA/4bpp Line:1 (Pitch 8) TMem:0x0100 Palette:0
         S: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
         T: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
LoadTLut: 0xf0af0000 0x07dec000 - Tile:7, (700,0) -> (891,0), Count 191
          Fmt is 16

////////////////////////////////////////////
// Works when * 4
SetTextureImage: Image: 0x001b2cd0 Fmt: CI/16 Width: 1
SetTile: 0xf5500000 0x07000000 - Tile:7  Fmt: CI/16bpp Line:0 (Pitch 0) TMem:0x0000 Palette:0
         S: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
         T: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
LoadBlock: 0xf3000000 0x073ff200 pad(0x00)
           Tile:7 (0,0) 1024 pixels DXT:0x0200 = 4 QWs => 16 pixels/line
          Offset: 0x001b2cd0

PipeSync

SetTile: 0xf5000300 0x07000000 - Tile:7  Fmt: RGBA/4bpp Line:1 (Pitch 8) TMem:0x0100 Palette:0
         S: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
         T: Clamp: Off Mirror:Off Mask:0x0 Shift:0x0
LoadTLut: 0xf0400c00 0x077fcc00 - Tile:7, (256,768) -> (511,768), Count 255
          Fmt is 16
*/

void RDP_GFX_LoadTLut(DWORD dwCmd0, DWORD dwCmd1)
{
	static const char *textluttype[4] = {"None", "?", "RGBA16", "IA16"};
	DWORD dwOffset;
	DWORD dwULS   = ((dwCmd0 >> 12) & 0xfff)/4;
	DWORD dwULT   = ((dwCmd0      ) & 0xfff)/4;
	DWORD dwTile  = ((dwCmd1 >> 24) & 0x07);
	DWORD dwLRS   = ((dwCmd1 >> 12) & 0xfff)/4;
	DWORD dwLRT   = ((dwCmd1      ) & 0xfff)/4;

	DWORD dwTLutFmt = (g_dwOtherModeH >> G_MDSFT_TEXTLUT)&0x3;
	DWORD dwCount = (dwLRS - dwULS);

	// Format is always 16bpp - RGBA16 or IA16:
	// I've no idea why these two are added - seems to work for 007!
	dwOffset = (dwULS + dwULT)*2;

	g_dwPalAddress = g_TI.dwAddr + dwOffset;
	g_dwPalSize = dwCount;

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		char str[300] = "";
		char item[300];
		WORD * pBase = (WORD *)(g_pu8RamBase + g_dwPalAddress);
		DWORD i;


		DL_PF("LoadTLut: 0x%08x 0x%08x - Tile:%d, (%d,%d) -> (%d,%d), Count %d",
			dwCmd0, dwCmd1, dwTile, dwULS, dwULT, dwLRS, dwLRT, dwCount);
		// This is sometimes wrong (in 007) tlut fmt is set after 
		// tlut load, but before tile load
		DL_PF("          Fmt is %s", textluttype[dwTLutFmt]);

		for (i = 0; i < g_dwPalSize; i++)
		{
			WORD wEntry = pBase[i ^ 0x1];

			if (i % 8 == 0)
			{
				DL_PF(str);

				// Clear
				wsprintf(str, "%03d: ", i);
			}

			wsprintf(item, "0x%04x (0x%08x) ", wEntry, Convert555ToRGBA(wEntry));
			lstrcat(str, item);
		}
		DL_PF(str);
	}

}



void RDP_GFX_RDPHalf_Cont(DWORD dwCmd0, DWORD dwCmd1) { /*DL_PF("RDPHalf_Cont: (Ignored)");*/ }
void RDP_GFX_RDPHalf_2(DWORD dwCmd0, DWORD dwCmd1)	{ /*DL_PF("RDPHalf_2: (Ignored)");*/ }
void RDP_GFX_RDPHalf_1(DWORD dwCmd0, DWORD dwCmd1)	{ /*DL_PF("RDPHalf_1: (Ignored)");*/ }
void RDP_GFX_RDPLoadSync(DWORD dwCmd0, DWORD dwCmd1)	{ /*DL_PF("LoadSync: (Ignored)");*/ }
void RDP_GFX_RDPPipeSync(DWORD dwCmd0, DWORD dwCmd1)
{
	/*DL_PF("PipeSync: (Ignored)");*/

	DL_PF("");
	DL_PF("PipeSync");
	DL_PF("");
}
void RDP_GFX_RDPTileSync(DWORD dwCmd0, DWORD dwCmd1)	{ /*DL_PF("TileSync: (Ignored)");*/ }

void RDP_GFX_RDPFullSync(DWORD dwCmd0, DWORD dwCmd1)
{ 
	Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_DP);
	AddCPUJob(CPU_CHECK_INTERRUPTS);
	
	/*DL_PF("FullSync: (Generating Interrupt)");*/
}


void RDP_GFX_DL(DWORD dwCmd0, DWORD dwCmd1)
{	
	DWORD dwPush = (dwCmd0 >> 16) & 0xFF;
	DWORD dwAddr = RDPSegAddr(dwCmd1);

	DL_PF("DisplayList: 0x%08x 0x%08x. Address=0x%08x Push: 0x%02x",
		dwCmd0, dwCmd1, RDPSegAddr(dwCmd1), dwPush);
	
	switch (dwPush)
	{
	case G_DL_PUSH:
		DL_PF("  Pushing DisplayList 0x%08x", dwAddr);
		g_dwPCStack.push_back( dwAddr );

		break;
	case G_DL_NOPUSH:
		DL_PF("  Jumping to DisplayList 0x%08x", dwAddr);
		g_dwPCStack[g_dwPCStack.size() - 1] = dwAddr;
		break;
	}
	DL_PF("");
	DL_PF("");
	DL_PF("#############################################");
}

void RDP_GFX_ZeldaDL(DWORD dwCmd0, DWORD dwCmd1)
{

	DWORD dwPush = (dwCmd0 >> 16) & 0xFF;
	DWORD dwAddr = RDPSegAddr(dwCmd1);

	DL_PF("ZeldaDL: 0x%08x 0x%08x Push:0x%02x Addr: 0x%08x", dwCmd0, dwCmd1, dwPush, dwAddr);
	
	switch (dwPush)
	{
	case G_DL_PUSH:
		DL_PF("  Pushing ZeldaDisplayList 0x%08x", dwAddr);
		g_dwPCStack.push_back( dwAddr );

		break;
	case G_DL_NOPUSH:
		DL_PF("  Jumping to ZeldaDisplayList 0x%08x", dwAddr);
		g_dwPCStack[g_dwPCStack.size() - 1] = dwAddr;
		break;
	}
	DL_PF("");
	DL_PF("");
	DL_PF("#############################################");


}



void RDP_GFX_EndDL(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("EndDisplayList: Returning from DisplayList");
	DL_PF("#############################################");
	DL_PF("");
	DL_PF("");

	if (g_dwPCStack.size() == 0)
	{
		DL_PF("EndDisplayList: Too many EndDL calls - ignoring");
		RDPHalt();			// Stop, damn it!
		return;
	}

	// Check if this is the lasy display list in the sequence
	if(g_dwPCStack.size() == 1)
	{
		RDPHalt();
		g_dwPCStack.clear();
		return;
	}

	// If we're here, then we musn't have finished with the display list yet
	g_dwPCStack.pop_back();
}


void RDP_GFX_ZeldaEndDL(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaEndDL: Returning from DisplayList");
	DL_PF("#############################################");
	DL_PF("");
	DL_PF("");

	if (g_dwPCStack.size() == 0)
	{
		DL_PF("ZeldaEndDL: Too many EndDL calls - ignoring");
		RDPHalt();			// Stop, damn it!
		return;
	}

	// Check if this is the lasy display list in the sequence
	if(g_dwPCStack.size() == 1)
	{
		RDPHalt();
		g_dwPCStack.clear();
		return;
	}

	// If we're here, then we musn't have finished with the display list yet
	g_dwPCStack.pop_back();
}



void RDP_GFX_MoveWord(DWORD dwCmd0, DWORD dwCmd1)
{
	// Type of movement is in low 8bits of cmd0.

	DWORD dwIndex = dwCmd0 & 0xFF;
	DWORD dwOffset = (dwCmd0 >> 8) & 0xFFFF;


	switch (dwIndex)
	{
		case G_MW_MATRIX:
			DL_PF("MoveWord: Type: G_MW_MATRIX");
			RDP_NOIMPL_WARN("G_MW_MATRIX Not Implemented");

			break;
		case G_MW_NUMLIGHT:
			//#define NUML(n)		(((n)+1)*32 + 0x80000000)
			{
				DWORD dwNumLights = ((dwCmd1-0x80000000)/32) - 1;
				DL_PF("MoveWord: Type: G_MW_NUMLIGHT: Val:%d", dwNumLights);

				g_dwAmbientLight = dwNumLights;
				g_Renderer->SetNumLights(dwNumLights);
			}
			break;
		case G_MW_CLIP:
			{
				switch (dwOffset)
				{
				case G_MWO_CLIP_RNX:
					DL_PF("MoveWord: Type: G_MW_CLIP  NegX: %d", (LONG)(SHORT)dwCmd1);
					break;
				case G_MWO_CLIP_RNY:
					DL_PF("MoveWord: Type: G_MW_CLIP  NegY: %d", (LONG)(SHORT)dwCmd1);
					break;
				case G_MWO_CLIP_RPX:
					DL_PF("MoveWord: Type: G_MW_CLIP  PosX: %d", (LONG)(SHORT)dwCmd1);
					break;
				case G_MWO_CLIP_RPY:
					DL_PF("MoveWord: Type: G_MW_CLIP  PosY: %d", (LONG)(SHORT)dwCmd1);
					break;
				default:
					DL_PF("MoveWord: Type: G_MW_CLIP  ?   : 0x%08x", dwCmd1);
					break;
				}

				//RDP_NOIMPL_WARN("G_MW_CLIP Not Implemented");
			}
			break;
		case G_MW_SEGMENT:
			{
				DWORD dwSegment = (dwOffset >> 2) & 0xF;
				DWORD dwBase = dwCmd1&0x00FFFFFF;
				DL_PF("MoveWord Segment: Seg[%d] = 0x%08x", dwSegment, dwBase);
				g_dwSegment[dwSegment] = dwBase;
			}
			break;
		case G_MW_FOG:
			{
				WORD wMult = (WORD)((dwCmd1 >> 16) & 0xFFFF);
				WORD wOff  = (WORD)((dwCmd1      ) & 0xFFFF);

				float fMult = (float)(short)wMult;
				float fOff = (float)(short)wOff;

				DL_PF("MoveWord: G_MW_FOG. Mult = 0x%04x, Off = 0x%04x", wMult, wOff);

				float fMin = 500.0f - ((128000.0f * fOff)/ (256.0f * fMult));
				float fMax = (128000.0f / fMult) + fMin;

				//	DBGConsole_Msg(0, "MoveWord: G_MW_FOG. Mult = 0x%04x, Off = 0x%04x", wMult, wOff);
				//	DBGConsole_Msg(0, "   => Min:%f, Max:%f", fMin/1000.0f, fMax/1000.0f);
				g_Renderer->SetFogRange(fMin/1000.0f, fMax/1000.0f);
			}
			//RDP_NOIMPL_WARN("G_MW_FOG Not Implemented");
			break;
		case G_MW_LIGHTCOL:
			{
				DWORD dwLight = dwOffset / 0x20;
				DWORD dwField = (dwOffset & 0x7);

				DL_PF("MoveWord: Type: G_MW_LIGHTCOL/0x%08x: 0x%08x", dwOffset, dwCmd1);

				switch (dwField)
				{
				case 0:
					//g_N64Lights[dwLight].dwRGBA = dwCmd1;
					// Light col, not the copy
					if (dwLight == g_dwAmbientLight)
					{
						g_Renderer->SetAmbientLight( (dwCmd1>>8) );
					}
					else
					{
						g_Renderer->SetLightCol(dwLight, dwCmd1);
					}
					break;

				case 4:
					break;

				default:
					//DBGConsole_Msg(0, "G_MW_LIGHTCOL with unknown offset 0x%08x", dwField);
					break;
				}
			}
			break;
		case G_MW_POINTS:
			DL_PF("MoveWord: Type: G_MW_POINTS");
			RDP_NOIMPL_WARN("G_MW_POINTS Not Implemented");
			break;
		case G_MW_PERSPNORM:
			DL_PF("MoveWord: Type: G_MW_PERSPNORM 0x%08x, 0x%08x", dwCmd0, dwCmd1);
			//RDP_NOIMPL_WARN("G_MW_PESPNORM Not Implemented");		// Used in Starfox - sets to 0xa
			//DBGConsole_Msg(0, "PerspNorm: 0x%08x", dwCmd1);		// 
			break;
		default:
			DL_PF("MoveWord: Type: Unknown");
			RDP_NOIMPL_WARN("Unknown MoveWord");
			break;
	}

}


//0016A710: DB020000 00000018 CMD Zelda_MOVEWORD  Mem[2][00]=00000018 Lightnum=0
//001889F0: DB020000 00000030 CMD Zelda_MOVEWORD  Mem[2][00]=00000030 Lightnum=2

void RDP_GFX_ZeldaMoveWord(DWORD dwCmd0, DWORD dwCmd1)
{

	DWORD dwType   = (dwCmd0 >> 16) & 0xFF;
	DWORD dwOffset = (dwCmd0      ) & 0xFFFF;

	DL_PF("ZeldaMoveWord: 0x%08x 0x%08x", dwCmd0, dwCmd1);

	if (dwType == G_MW_NUMLIGHT)
	{
		// Lightnum
		// dwCmd1:
		// 0x18 = 24 = 0 lights
		// 0x30 = 48 = 2 lights

		DWORD dwNumLights = (dwCmd1/12) - 2;
		DL_PF("     Type: G_MW_NUMLIGHT: Val:%d", dwNumLights);

		g_dwAmbientLight = dwNumLights;
		g_Renderer->SetNumLights(dwNumLights);
	}
	else if (dwType == G_MW_SEGMENT)
	{
		DWORD dwSeg     = dwOffset / 4;
		DWORD dwAddress = dwCmd1 & 0x00FFFFFF;			// Hack - convert to physical

		DL_PF("      Segment[%d] = 0x%08x",
			dwSeg, dwAddress);

		g_dwSegment[dwSeg] = dwAddress;

	}
	else
	{
		DL_PF("      Ignored!!", dwCmd0, dwCmd1);

	}
}

void RDP_GFX_MoveMem(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwType    = (dwCmd0>>16)&0xFF;
	DWORD dwLength  = (dwCmd0)&0xFFFF;
	DWORD dwAddress = RDPSegAddr(dwCmd1);

	switch (dwType) {
		case G_MV_VIEWPORT:
			{
				DL_PF("MoveMem: Type: G_MV_VIEWPORT. Address: 0x%08x, Length: 0x%04x", dwAddress, dwLength);
				RDP_MoveMemViewport(dwAddress);
			}
			break;
		case G_MV_LOOKATY:
			DL_PF("MoveMem: Type: G_MV_LOOKATY");
			//RDP_NOIMPL_WARN("G_MV_LOOKATY Not Implemented");
			break;
		case G_MV_LOOKATX:
			DL_PF("MoveMem: Type: G_MV_LOOKATX");
			//RDP_NOIMPL_WARN("G_MV_LOOKATX Not Implemented");
			break;
		case G_MV_L0:
		case G_MV_L1:
		case G_MV_L2:
		case G_MV_L3:
		case G_MV_L4:
		case G_MV_L5:
		case G_MV_L6:
		case G_MV_L7:
			{
				DL_PF("MoveMem: Type: G_MV_L*");

				// Ensure dwType == G_MV_L0 -- G_MV_L7
				//	if (dwType < G_MV_L0 || dwType > G_MV_L7 || ((dwType & 0x1) != 0))
				//		break;

				DWORD dwLight = (dwType-G_MV_L0)/2;

				DL_PF("MoveMem Light%d: Length:0x%04x, Address: 0x%08x", dwLight, dwLength, dwAddress);

				RDP_MoveMemLight(dwLight, dwAddress);
			}
			break;
		case G_MV_TXTATT:
			DL_PF("MoveMem: Type: G_MV_TXTATT");
			RDP_NOIMPL_WARN("G_MV_TXTATT Not Implemented");
			break;
		case G_MV_MATRIX_1:
			DL_PF("MoveMem: Type: G_MV_MATRIX_1");
			RDP_NOIMPL_WARN("G_MV_MATRIX_1 Not Implemented");
			break;
		case G_MV_MATRIX_2:
			DL_PF("MoveMem: Type: G_MV_MATRIX_2");
			RDP_NOIMPL_WARN("G_MV_MATRIX_2 Not Implemented");
			break;
		case G_MV_MATRIX_3:
			DL_PF("MoveMem: Type: G_MV_MATRIX_3");
			RDP_NOIMPL_WARN("G_MV_MATRIX_3 Not Implemented");
			break;
		case G_MV_MATRIX_4:
			DL_PF("MoveMem: Type: G_MV_MATRIX_4");
			RDP_NOIMPL_WARN("G_MV_MATRIX_4 Not Implemented");
			break;
		default:
			DL_PF("MoveMem: 0x%08x 0x%08x Type: Unknown", dwCmd0, dwCmd1);
			RDP_NOIMPL_WARN("Unknown Move Type");
			break;
	}
}
/*

001889F8: DC08060A 80188708 CMD Zelda_MOVEMEM  Movemem[0806] <- 80188708
!light 0 color 0.12 0.16 0.35 dir 0.01 0.00 0.00 0.00 (2 lights) [ 1E285A00 1E285A00 01000000 00000000 ]
data(00188708): 1E285A00 1E285A00 01000000 00000000 
00188A00: DC08090A 80188718 CMD Zelda_MOVEMEM  Movemem[0809] <- 80188718
!light 1 color 0.23 0.25 0.30 dir 0.01 0.00 0.00 0.00 (2 lights) [ 3C404E00 3C404E00 01000000 00000000 ]
data(00188718): 3C404E00 3C404E00 01000000 00000000 
00188A08: DC080C0A 80188700 CMD Zelda_MOVEMEM  Movemem[080C] <- 80188700
!light 2 color 0.17 0.16 0.26 dir 0.23 0.31 0.70 0.00 (2 lights) [ 2C294300 2C294300 1E285A00 1E285A00 ]
*/
/*
ZeldaMoveMem: 0xdc080008 0x801984d8
SetScissor: x0=416 y0=72 x1=563 y1=312 mode=0
// Mtx
ZeldaMoveWord:0xdb0e0000 0x00000041 Ignored
ZeldaMoveMem: 0xdc08000a 0x80198538
ZeldaMoveMem: 0xdc08030a 0x80198548

ZeldeMoveMem: Unknown Type. 0xdc08000a 0x80198518
ZeldeMoveMem: Unknown Type. 0xdc08030a 0x80198528
ZeldeMoveMem: Unknown Type. 0xdc08000a 0x80198538
ZeldeMoveMem: Unknown Type. 0xdc08030a 0x80198548
ZeldeMoveMem: Unknown Type. 0xdc08000a 0x80198518
ZeldeMoveMem: Unknown Type. 0xdc08030a 0x80198528
ZeldeMoveMem: Unknown Type. 0xdc08000a 0x80198538
ZeldeMoveMem: Unknown Type. 0xdc08030a 0x80198548

*/
void RDP_GFX_ZeldaMoveMem(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaMoveMem: 0x%08x 0x%08x", dwCmd0, dwCmd1);

	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwOffset = (dwCmd0 >> 8) & 0xFFFF;
	DWORD dwUnk    = (dwCmd0     ) & 0xFF;

	switch (dwOffset)
	{
	case 0x0800:
		if (dwUnk == 0x08)
		{
			RDP_MoveMemViewport(dwAddress);
		}
		break;
	case 0x0806:
	case 0x0809:
	case 0x080C:
		{
			DWORD dwLight = (dwOffset - 0x0806)/3;
			DL_PF("              Light %d:",dwLight);
			RDP_MoveMemLight(dwLight, dwAddress);

		}
		break;

	default:
		{
			//DBGConsole_Msg(0, "ZeldeMoveMem: Unknown Type. 0x%08x 0x%08x", dwCmd0, dwCmd1);
			/*s8 * pcBase = g_ps8RamBase + dwAddress;
			DWORD * pdwBase = (DWORD *)pcBase;
			LONG i;
			
			for (i = 0; i < 16; i+=4)
			{
				DL_PF("   0x%08x", pdwBase[i/4]);
			}*/
		}
		break;
	}


}

void RDP_GFX_Mtx(DWORD dwCmd0, DWORD dwCmd1)
{	
	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwCommand = (dwCmd0>>16)&0xFF;
	DWORD dwLength  = (dwCmd0)    &0xFFFF;

	DL_PF("Matrix: Command: %s %s %s Length %d Address 0x%08x",
		(dwCommand & G_MTX_PROJECTION) ? "Projection" : "ModelView",
		(dwCommand & G_MTX_LOAD) ? "Load" : "Mul",	
		(dwCommand & G_MTX_PUSH) ? "Push" : "NoPush",
		dwLength, dwAddress);

	if (dwAddress + 64 > g_dwRamSize)
	{
		DBGConsole_Msg(0, "Mtx: Address invalid (0x%08x)", dwAddress);
		return;
	}

	// Load matrix from dwAddress
	D3DXMATRIX mat;
	DWORD dwI;
	DWORD dwJ;
	const float fRecip = 1.0f / 65536.0f;
	
	for (dwI = 0; dwI < 4; dwI++) {
		for (dwJ = 0; dwJ < 4; dwJ++) {

			SHORT nDataHi = *(SHORT *)(g_pu8RamBase + ((dwAddress+(dwI<<3)+(dwJ<<1)     )^0x2));
			WORD  nDataLo = *(WORD  *)(g_pu8RamBase + ((dwAddress+(dwI<<3)+(dwJ<<1) + 32)^0x2));

			mat.m[dwI][dwJ] = (float)(((LONG)nDataHi<<16) | (nDataLo))*fRecip;
		}
	}

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DL_PF(
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			mat.m[0][0], mat.m[0][1], mat.m[0][2], mat.m[0][3],
			mat.m[1][0], mat.m[1][1], mat.m[1][2], mat.m[1][3],
			mat.m[2][0], mat.m[2][1], mat.m[2][2], mat.m[2][3],
			mat.m[3][0], mat.m[3][1], mat.m[3][2], mat.m[3][3]);
	}

	LONG nLoadCommand = dwCommand & G_MTX_LOAD ? D3DRENDER_LOAD : D3DRENDER_MUL;
	BOOL bPush = dwCommand & G_MTX_PUSH ? TRUE : FALSE;


	if (dwCommand & G_MTX_PROJECTION)
	{
		// So far only Extreme-G seems to Push/Pop projection matrices	
		g_Renderer->SetProjection(mat, bPush, nLoadCommand);
	}
	else
	{
		g_Renderer->SetWorldView(mat, bPush, nLoadCommand);
	}
}

void RDP_GFX_PopMtx(DWORD dwCmd0, DWORD dwCmd1)
{
	BYTE nCommand = (BYTE)(dwCmd1 & 0xFF);

	DL_PF("PopMatrix: 0x%08x 0x%08x Command: 0x%02x (%s)",
		dwCmd0, dwCmd1, nCommand, 
		(nCommand & G_MTX_PROJECTION) ? "Projection" : "ModelView");

	// Do any of the other bits do anything?
	// So far only Extreme-G seems to Push/Pop projection matrices

	if (nCommand & G_MTX_PROJECTION)
	{
		g_Renderer->PopProjection();
	}
	else
	{
		g_Renderer->PopWorldView();
	}
}

/*
0016A748: DA380007 801764C8 CMD Zelda_LOADMTX  {Matrix} at 001764C8  Load:Prj 
Load matrix(1):
0016A750: DA380005 80176488 CMD Zelda_LOADMTX  {Matrix} at 00176488  Mul:Prj 
Mul matrix(1):
0016A760: DA380003 80176448 CMD Zelda_LOADMTX  {Matrix} at 00176448  Load:Mod 
Load matrix(0):
0016AA70: DA380000 80175F88 CMD Zelda_LOADMTX  {Matrix} at 00175F88  Mul:Mod Push 
Mul matrix(0): (push)


#define	G_MTX_MODELVIEW		0x00
#define	G_MTX_PROJECTION	0x01
#define	G_MTX_MUL			0x00
#define	G_MTX_LOAD			0x02
#define G_MTX_NOPUSH		0x00
#define G_MTX_PUSH			0x04
*/
#define G_ZMTX_MODELVIEW	0x00
#define G_ZMTX_PROJECTION	0x04
#define G_ZMTX_MUL			0x00
#define G_ZMTX_LOAD			0x02
#define G_ZMTX_PUSH			0x00
#define G_ZMTX_NOPUSH		0x01

void RDP_GFX_ZeldaMtx(DWORD dwCmd0, DWORD dwCmd1)
{	
	D3DXMATRIX mat;
	DWORD dwI;
	DWORD dwJ;
	const float fRecip = 1.0f / 65535.0f;
	DWORD dwAddress = RDPSegAddr(dwCmd1);

	// THESE ARE SWAPPED OVER FROM NORMAL MTX
	// Not sure if this is right!!!!
	DWORD dwCommand = (dwCmd0)&0xFF;
	DWORD dwLength  = (dwCmd0 >> 8)&0xFFFF;

	DL_PF("ZeldaMatrix: 0x%08x 0x%08x Command: %s %s %s Length %d Address 0x%08x",
		dwCmd0, dwCmd1,
		(dwCommand & G_ZMTX_PROJECTION) ? "Projection" : "ModelView",
		(dwCommand & G_ZMTX_LOAD) ? "Load" : "Mul",	
		(dwCommand & G_ZMTX_NOPUSH) ? "NoPush" : "Push",
		dwLength, dwAddress);

	if (dwAddress + 64 > g_dwRamSize)
	{
		DBGConsole_Msg(0, "ZeldaMtx: Address invalid (0x%08x)", dwAddress);
		return;
	}

	// Load matrix from dwAddress	
	for (dwI = 0; dwI < 4; dwI++)
	{
		for (dwJ = 0; dwJ < 4; dwJ++) 
		{
			SHORT nDataHi = *(SHORT *)(g_pu8RamBase + ((dwAddress+(dwI<<3)+(dwJ<<1)     )^0x2));
			WORD  nDataLo = *(WORD  *)(g_pu8RamBase + ((dwAddress+(dwI<<3)+(dwJ<<1) + 32)^0x2));

			mat.m[dwI][dwJ] = (float)(((LONG)nDataHi<<16) | (nDataLo))*fRecip;
		}
	}

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DL_PF(
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\r\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\r\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\r\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\r\n",
			mat.m[0][0], mat.m[0][1], mat.m[0][2], mat.m[0][3],
			mat.m[1][0], mat.m[1][1], mat.m[1][2], mat.m[1][3],
			mat.m[2][0], mat.m[2][1], mat.m[2][2], mat.m[2][3],
			mat.m[3][0], mat.m[3][1], mat.m[3][2], mat.m[3][3]);
	}

	LONG nLoadCommand = dwCommand & G_ZMTX_LOAD ? D3DRENDER_LOAD : D3DRENDER_MUL;
	BOOL bPush = dwCommand & G_ZMTX_NOPUSH ? FALSE : TRUE;


	if (dwCommand & G_ZMTX_PROJECTION)
	{
		// So far only Extreme-G seems to Push/Pop projection matrices	
		g_Renderer->SetProjection(mat, bPush, nLoadCommand);
	}
	else
	{
		g_Renderer->SetWorldView(mat, bPush, nLoadCommand);
	}
}

void RDP_GFX_ZeldaPopMtx(DWORD dwCmd0, DWORD dwCmd1)
{
	BYTE nCommand = (BYTE)(dwCmd0 & 0xFF);

	DL_PF("ZeldaPopMatrix: 0x%08x 0x%08x Command: 0x%02x (%s)",
		dwCmd0, dwCmd1, nCommand, 
		(nCommand & G_ZMTX_PROJECTION) ? "Projection" : "ModelView");


/*	if (nCommand & G_ZMTX_PROJECTION)
	{
		g_Renderer->PopProjection();
	}
	else*/
	{
		g_Renderer->PopWorldView();
	}

}

/*	
cc = cmd
ff = first
ll = len
          ccffll??
802C80C0: 01004008 01001C00 CMD Zelda_LOADVTX  Vertex 00..03 at 802C7480
80317030: 01020040 08000EB0 CMD Zelda_LOADVTX  Vertex 00..31 at 802048D0
80317140: 01010020 08001470 CMD Zelda_LOADVTX  Vertex 00..15 at 80204E90
80317228: 0100F01E 08001670 CMD Zelda_LOADVTX  Vertex 00..14 at 80205090
80315FE8: 0100D01A 0600A580 CMD Zelda_LOADVTX  Vertex 00..12 at 803158D0
80316060: 0101B036 0600A650 CMD Zelda_LOADVTX  Vertex 00..26 at 803159A0

802F9C48: 0100E024 0602D9E0 CMD Zelda_LOADVTX  Vertex 04..17 at 802F7150

*/
void RDP_GFX_Vtx_Zelda(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwVEnd    = ((dwCmd0   )&0xFFF)/2;
	DWORD dwN      = ((dwCmd0>>12)&0xFFF);

	DWORD dwV0		= dwVEnd - dwN;


	DL_PF("ZeldaVertex: 0x%08x 0x%08x ", dwCmd0, dwCmd1);
	DL_PF("             Address 0x%08x, vEnd: %d, v0: %d, Num: %d", dwAddress, dwVEnd, dwV0, dwN);

	if (dwV0 >= 32)
		dwV0 = 31;
	
	if ((dwV0 + dwN) > 32)
	{
		DL_PF("        Warning, attempting to load into invalid vertex positions");
		DBGConsole_Msg(0, "Warning, attempting to load into invalid vertex positions");
		dwN = 32 - dwV0;
	}

	// Check that address is valid...
	if ((dwAddress + (dwN*16)) > g_dwRamSize)
	{
		DBGConsole_Msg(0, "SetNewVertexInfo: Address out of range (0x%08x)", dwAddress);
	}
	else
	{
		g_Renderer->SetNewVertexInfo(dwAddress, dwV0, dwN);

		g_dwNumVertices += dwN;

		RDP_GFX_DumpVtxInfo(dwAddress, dwV0, dwN);
	}

}


void RDP_GFX_Vtx_Mario(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwV0 =  (dwCmd0>>16)&0x0F;
	DWORD dwN  = ((dwCmd0>>20)&0x0F)+1;
	DWORD dwLength = (dwCmd0)&0xFFFF;

	DL_PF("Vertex: 0x%08x 0x%08x ", dwCmd0, dwCmd1);
	DL_PF("        Address 0x%08x, v0: %d, Num: %d, Length: 0x%04x", dwAddress, dwV0, dwN, dwLength);

	if (dwV0 >= 32)
		dwV0 = 31;
	
	if ((dwV0 + dwN) > 32)
	{
		DL_PF("        Warning, attempting to load into invalid vertex positions");
		DBGConsole_Msg(0, "Warning, attempting to load into invalid vertex positions");
		dwN = 32 - dwV0;
	}

	// Check that address is valid...
	if ((dwAddress + (dwN*16)) > g_dwRamSize)
	{
		DBGConsole_Msg(0, "SetNewVertexInfo: Address out of range (0x%08x)", dwAddress);
	}
	else
	{
		g_Renderer->SetNewVertexInfo(dwAddress, dwV0, dwN);

		g_dwNumVertices += dwN;

		RDP_GFX_DumpVtxInfo(dwAddress, dwV0, dwN);
	}
}

void RDP_GFX_Vtx_MarioKart(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwLength = (dwCmd0)&0xFFFF;

	DWORD dwN= (dwLength + 1) / 0x410;
	DWORD dwV0 = ((dwCmd0>>16)&0x3f)>>1;

	DL_PF("Vertex: 0x%08x 0x%08x ", dwCmd0, dwCmd1);
	DL_PF("        Address 0x%08x, v0: %d, Num: %d, Length: 0x%04x", dwAddress, dwV0, dwN, dwLength);

	if (dwV0 >= 32)
		dwV0 = 31;
	
	if ((dwV0 + dwN) > 32)
	{
		DL_PF("        Warning, attempting to load into invalid vertex positions");
		dwN = 32 - dwV0;
	}

	g_Renderer->SetNewVertexInfo(dwAddress, dwV0, dwN);

	g_dwNumVertices += dwN;

	RDP_GFX_DumpVtxInfo(dwAddress, dwV0, dwN);
}

void RDP_GFX_Vtx_WaveraceUS(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAddress = RDPSegAddr(dwCmd1);
	DWORD dwLength = (dwCmd0)&0xFFFF;

	DWORD dwN= (dwLength + 1) / 0x210;
	//DWORD dwV0 = ((dwCmd0>>16)&0x3f)/5;
	DWORD dwV0 = ((dwCmd0>>16)&0xFF)/5;

	DL_PF("Vertex: 0x%08x 0x%08x ", dwCmd0, dwCmd1);
	DL_PF("        Address 0x%08x, v0: %d, Num: %d, Length: 0x%04x", dwAddress, dwV0, dwN, dwLength);

	if (dwV0 >= 32)
		dwV0 = 31;
	
	if ((dwV0 + dwN) > 32)
	{
		DL_PF("        Warning, attempting to load into invalid vertex positions");
		DBGConsole_Msg(0, "Warning, attempting to load into invalid vertex positions");
		dwN = 32 - dwV0;
	}

	g_Renderer->SetNewVertexInfo(dwAddress, dwV0, dwN);

	g_dwNumVertices += dwN;

	RDP_GFX_DumpVtxInfo(dwAddress, dwV0, dwN);
}



void RDP_GFX_DumpVtxInfo(DWORD dwAddress, DWORD dwV0, DWORD dwN)
{
	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		char *pcSrc = (char *)(g_pu8RamBase + dwAddress);
		short *psSrc = (short *)(g_pu8RamBase + dwAddress);

		for (DWORD dwV = dwV0; dwV < dwV0 + dwN; dwV++)
		{
			float x = (float)psSrc[0^0x1];
			float y = (float)psSrc[1^0x1];
			float z = (float)psSrc[2^0x1];

			WORD wFlags = g_Renderer->m_dwVecFlags[dwV]; //(WORD)psSrc[3^0x1];

			BYTE a = pcSrc[12^0x3];
			BYTE b = pcSrc[13^0x3];
			BYTE c = pcSrc[14^0x3];
			BYTE d = pcSrc[15^0x3];
			
			LONG nTU = (LONG)(SHORT)(psSrc[4^0x1]<<4);
			LONG nTV = (LONG)(SHORT)(psSrc[5^0x1]<<4);


			float tu = (float)(nTU>>4);
			float tv = (float)(nTV>>4);

			D3DXVECTOR3 & t = g_Renderer->m_vecProjected[dwV];

			psSrc += 8;			// Increase by 16 bytes
			pcSrc += 16;

			DL_PF(" #%02d Flags: 0x%04x Pos: {% 6f,% 6f,% 6f} Tex: {%+7.2f,%+7.2f}, Extra: %02x %02x %02x %02x (transf: {% 6f,% 6f,% 6f})",
				dwV, wFlags, x, y, z, tu, tv, a, b, c, d, t.x, t.y, t.z );
		}
	}

}



/*
Zelda: Reserved0(0x022516ff 0x002814ff)
Zelda: Reserved0(0x022d18ff 0x004c22ff)
Zelda: Reserved0(0x024d2bff 0x187347ff)
SetNewVertexInfo: Address out of range (0x00c0f53f)
Zelda: Reserved0(0x022f1dff 0x085031ff)
SetNewVertexInfo: Address out of range (0x009c9b1f)
SetNewVertexInfo: Address out of range (0x00cc1e3f)
Zelda: Reserved0(0x02301cff 0x052618ff)
Zelda: Reserved0(0x020d1de2 0x0f1d3fff)
Warning, attempting to load into invalid vertex positions
Warning, attempting to load into invalid vertex positions
Warning, attempting to load into invalid vertex positions
Zelda: Reserved0(0x020815ff 0x05142eff)
Exception processing 0x05050300 0x00000000*/

void RDP_GFX_ZeldaTri1(DWORD dwCmd0, DWORD dwCmd1)
{
	// While the next command pair is Tri1, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];
	DWORD * pCmdBase = (DWORD *)(g_pu8RamBase + dwPC);

	BOOL bTrisAdded = FALSE;
	
	do
	{
		//DWORD dwFlag = (dwCmd1>>24)&0xFF;
		// Vertex indices are multiplied by 10 for Mario64, by 2 for MarioKart
		DWORD dwV2 = ((dwCmd0>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1 = ((dwCmd0>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV0 = ((dwCmd0   )&0xFF)/g_dwVertexMult;


		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			DL_PF("ZeldaTri1: 0x%08x 0x%08x %d,%d,%d", dwCmd0, dwCmd1, dwV0, dwV1, dwV2);
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);
			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
		{
			DL_PF("ZeldaTri1: 0x%08x 0x%08x %d,%d,%d (clipped)", dwCmd0, dwCmd1, dwV0, dwV1, dwV2);
			g_dwNumTrisClipped++;
		}

		dwCmd0			= *pCmdBase++;
		dwCmd1			= *pCmdBase++;
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_ZELDATRI1);
	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;


	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}
}


//802C80C8: 06000204 00000406 CMD Zelda_TRI2  (2,1,0)
//00284030: 07002026 00002624 CMD Zelda_TRI3  (19,16,0)

void RDP_GFX_ZeldaTri2(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaTri2: 0x%08x 0x%08x", dwCmd0, dwCmd1);

	// While the next command pair is Tri2, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	BOOL bTrisAdded = FALSE;

	do {
		DWORD dwV2 = ((dwCmd1>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1 = ((dwCmd1>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV0 = ((dwCmd1    )&0xFF)/g_dwVertexMult;

		DWORD dwV5 = ((dwCmd0>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV4 = ((dwCmd0>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV3 = ((dwCmd0    )&0xFF)/g_dwVertexMult;

		DL_PF("ZeldaTri2: 0x%08x 0x%08x", dwCmd0, dwCmd1);
		DL_PF("           V0: %d, V1: %d, V2: %d", dwV0, dwV1, dwV2);
		DL_PF("           V3: %d, V4: %d, V5: %d", dwV3, dwV4, dwV5);

		// Do first tri
		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;

		// Do second tri
		if (g_Renderer->TestTri(dwV3, dwV4, dwV5))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV3, dwV4, dwV5);
			g_dwNumTrisRendered++;
		}
		else
		{
			g_dwNumTrisClipped++;
		}
		
		dwCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
		dwCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_ZELDATRI2);


	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;


	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}
}

//00284030: 07002026 00002624 CMD Zelda_TRI3  (19,16,0)
void RDP_GFX_ZeldaTri3(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("ZeldaTri3: 0x%08x 0x%08x", dwCmd0, dwCmd1);

	// While the next command pair is Tri2, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	BOOL bTrisAdded = FALSE;

	do {
		// Vertex indices are multiplied by 10 for Mario64, by 2 for MarioKart
		DWORD dwV2 = ((dwCmd1>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1 = ((dwCmd1>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV0 = ((dwCmd1    )&0xFF)/g_dwVertexMult;

		DWORD dwV5 = ((dwCmd0>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV4 = ((dwCmd0>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV3 = ((dwCmd0    )&0xFF)/g_dwVertexMult;

		DL_PF("ZeldaTri3: 0x%08x 0x%08x", dwCmd0, dwCmd1);
		DL_PF("           V0: %d, V1: %d, V2: %d", dwV0, dwV1, dwV2);
		DL_PF("           V3: %d, V4: %d, V5: %d", dwV3, dwV4, dwV5);

		// Do first tri
		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);
			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;

		// Do second tri
		if (g_Renderer->TestTri(dwV3, dwV4, dwV5))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);
			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV3, dwV4, dwV5);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;
		
		dwCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
		dwCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_ZELDATRI3);


	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;


	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}
}


void RDP_GFX_Tri1(DWORD dwCmd0, DWORD dwCmd1)
{
	// While the next command pair is Tri1, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];
	DWORD * pCmdBase = (DWORD *)(g_pu8RamBase + dwPC);
	
	BOOL bTrisAdded = FALSE;

	do
	{
		//DWORD dwFlag = (dwCmd1>>24)&0xFF;
		// Vertex indices are multiplied by 10 for Mario64, by 2 for MarioKart
		DWORD dwV0 = ((dwCmd1>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1 = ((dwCmd1>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV2 = ((dwCmd1    )&0xFF)/g_dwVertexMult;

		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			DL_PF("Tri1: 0x%08x 0x%08x %d,%d,%d", dwCmd0, dwCmd1, dwV0, dwV1, dwV2);

			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
		{
			DL_PF("Tri1: 0x%08x 0x%08x %d,%d,%d (clipped)", dwCmd0, dwCmd1, dwV0, dwV1, dwV2);
			g_dwNumTrisClipped++;
		}

		dwCmd0			= *pCmdBase++;
		dwCmd1			= *pCmdBase++;
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_TRI1);
	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;

	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}

}

void RDP_GFX_Line3D(DWORD dwCmd0, DWORD dwCmd1)
{
	// While the next command pair is Tri1, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	BOOL bTrisAdded = FALSE;


	do {
		DWORD dwV3   = ((dwCmd1>>24)&0xFF)/g_dwVertexMult;		
		DWORD dwV0   = ((dwCmd1>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1	 = ((dwCmd1>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV2   = ((dwCmd1    )&0xFF)/g_dwVertexMult;

		DL_PF("Line3D: V0: %d, V1: %d, V2: %d, V3: %d", dwV0, dwV1, dwV2, dwV3);

		// Do first tri
		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;

		// Do second tri
		if (g_Renderer->TestTri(dwV2, dwV3, dwV0))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV2, dwV3, dwV0);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;


		dwCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
		dwCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_LINE3D);

	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;

	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}	
}


void RDP_GFX_Tri2_Goldeneye(DWORD dwCmd0, DWORD dwCmd1)
{
	// While the next command pair is Tri2, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	BOOL bTrisAdded = FALSE;

	do {
		// Vertex indices are exact
		DWORD dwB = (dwCmd1    ) & 0xF;
		DWORD dwA = (dwCmd1>> 4) & 0xF;
		DWORD dwE = (dwCmd1>> 8) & 0xF;
		DWORD dwD = (dwCmd1>>12) & 0xF;
		DWORD dwH = (dwCmd1>>16) & 0xF;
		DWORD dwG = (dwCmd1>>20) & 0xF;
		DWORD dwK = (dwCmd1>>24) & 0xF;
		DWORD dwJ = (dwCmd1>>28) & 0xF;

		DWORD dwC = (dwCmd0    ) & 0xF;
		DWORD dwF = (dwCmd0>> 4) & 0xF;
		DWORD dwI = (dwCmd0>> 8) & 0xF;
		DWORD dwL = (dwCmd0>>12) & 0xF;

		DWORD dwFlag = (dwCmd0>>16)&0xFF;

		BOOL bVisible;

		DL_PF("Tri2: 0x%08x 0x%08x Flag: 0x%02x", dwCmd0, dwCmd1, dwFlag);

		// Don't check the first two tris for nullity
		bVisible = g_Renderer->TestTri(dwA, dwC, dwB);
		DL_PF("     (%d, %d, %d) %s", dwA, dwB, dwC, bVisible ? "": "(clipped)");
		if (bVisible)
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwA, dwC, dwB);
			g_dwNumTrisRendered++;
		}
		else
		{
			g_dwNumTrisClipped++;
		}

		
		bVisible = g_Renderer->TestTri(dwD, dwF, dwE);
		DL_PF("     (%d, %d, %d) %s", dwD, dwE, dwF, bVisible ? "": "(clipped)");
		if (bVisible)
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwD, dwF, dwE);
			g_dwNumTrisRendered++;
		}
		else
		{
			g_dwNumTrisClipped++;
		}


		if (dwG != dwI && dwI != dwH && dwH != dwG)
		{
			bVisible = g_Renderer->TestTri(dwG, dwI, dwH);
			DL_PF("     (%d, %d, %d) %s", dwG, dwH, dwI, bVisible ? "": "(clipped)");
			if (bVisible)
			{
				if (!bTrisAdded && g_bTextureEnable)
					RDP_EnableTexturing(g_dwTextureTile);

				bTrisAdded = TRUE;
				g_Renderer->AddTri(dwG, dwI, dwH);
				g_dwNumTrisRendered++;
			}
			else
			{
				g_dwNumTrisClipped++;
			}
		}

		if (dwJ != dwL && dwL != dwK && dwK != dwJ)
		{
			bVisible = g_Renderer->TestTri(dwJ, dwL, dwK);
			DL_PF("     (%d, %d, %d) %s", dwJ, dwK, dwL, bVisible ? "": "(clipped)");
			if (bVisible)
			{
				if (!bTrisAdded && g_bTextureEnable)
					RDP_EnableTexturing(g_dwTextureTile);

				bTrisAdded = TRUE;
				g_Renderer->AddTri(dwJ, dwL, dwK);
				g_dwNumTrisRendered++;
			}
			else
			{
				g_dwNumTrisClipped++;
			}
		}
		
		dwCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
		dwCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_TRI2);


	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;


	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}		
	else
	{
		g_dwNumTexturesIgnored++;
	}
}



/* Mariokart etc*/
void RDP_GFX_Tri2_Mariokart(DWORD dwCmd0, DWORD dwCmd1)
{

	// While the next command pair is Tri2, add vertices
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	BOOL bTrisAdded = FALSE;

	do {
		// Vertex indices are multiplied by 10 for Mario64, by 2 for MarioKart
		DWORD dwV0 = ((dwCmd1>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV1 = ((dwCmd1>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV2 = ((dwCmd1    )&0xFF)/g_dwVertexMult;

		DWORD dwV3 = ((dwCmd0>>16)&0xFF)/g_dwVertexMult;
		DWORD dwV4 = ((dwCmd0>>8 )&0xFF)/g_dwVertexMult;
		DWORD dwV5 = ((dwCmd0    )&0xFF)/g_dwVertexMult;

		DL_PF("Tri2: 0x%08x 0x%08x", dwCmd0, dwCmd1);
		DL_PF("      V0: %d, V1: %d, V2: %d", dwV0, dwV1, dwV2);
		DL_PF("      V3: %d, V4: %d, V5: %d", dwV3, dwV4, dwV5);

		// Do first tri
		if (g_Renderer->TestTri(dwV0, dwV1, dwV2))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV0, dwV1, dwV2);
			g_dwNumTrisRendered++;
		}
		else
			g_dwNumTrisClipped++;

		// Do second tri
		if (g_Renderer->TestTri(dwV3, dwV4, dwV5))
		{
			if (!bTrisAdded && g_bTextureEnable)
				RDP_EnableTexturing(g_dwTextureTile);

			bTrisAdded = TRUE;
			g_Renderer->AddTri(dwV3, dwV4, dwV5);
			g_dwNumTrisRendered++;
		}
		else
		{
			g_dwNumTrisClipped++;
		}
		
		dwCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
		dwCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);
		dwPC += 8;

	} while ((dwCmd0>>24) == (BYTE)G_TRI2);


	g_dwPCStack[g_dwPCStack.size() - 1] = dwPC-8;


	if (bTrisAdded)	
	{
		g_Renderer->FlushTris();
	}
	else
	{
		g_dwNumTexturesIgnored++;
	}
}


void RDP_GFX_SetScissor(DWORD dwCmd0, DWORD dwCmd1)
{
	// The coords are all in 8:2 fixed point
	g_Scissor.x0   = (dwCmd0>>12)&0xFFF;
	g_Scissor.y0   = (dwCmd0>>0 )&0xFFF;
	g_Scissor.mode = (dwCmd1>>24)&0x03;
	g_Scissor.x1   = (dwCmd1>>12)&0xFFF;
	g_Scissor.y1   = (dwCmd1>>0 )&0xFFF;

	DL_PF("SetScissor: x0=%d y0=%d x1=%d y1=%d mode=%d",
		g_Scissor.x0/4, g_Scissor.y0/4,
		g_Scissor.x1/4, g_Scissor.y1/4,
		g_Scissor.mode);


	// Set the cliprect now...

	//RDP_NOIMPL_WARN("GFX_SetScissor Not Implemented");
}


void RDP_GFX_SetTile(DWORD dwCmd0, DWORD dwCmd1)
{
	char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
	char *pszImgSize[4]   = {"4bpp", "8bpp", "16bpp", "32bpp"};
	char *pszOnOff[2]     = {"Off", "On"};

	DWORD dwTile		= (dwCmd1>>24)&0x7;


	g_Tiles[dwTile].dwFormat	= (dwCmd0>>21)&0x7;
	g_Tiles[dwTile].dwSize		= (dwCmd0>>19)&0x3;
	g_Tiles[dwTile].dwLine		= (dwCmd0>>9 )&0x1FF;
	g_Tiles[dwTile].dwTMem		= (dwCmd0    )&0x1FF;

	g_Tiles[dwTile].dwPalette	= (dwCmd1>>20)&0x0F;
	g_Tiles[dwTile].bClampT		= (dwCmd1>>19)&0x01;
	g_Tiles[dwTile].bMirrorT	= (dwCmd1>>18)&0x01;
	g_Tiles[dwTile].dwMaskT		= (dwCmd1>>14)&0x0F;
	g_Tiles[dwTile].dwShiftT	= (dwCmd1>>10)&0x0F;		// LOD stuff
	g_Tiles[dwTile].bClampS		= (dwCmd1>>9 )&0x01;
	g_Tiles[dwTile].bMirrorS	= (dwCmd1>>8 )&0x01;
	g_Tiles[dwTile].dwMaskS		= (dwCmd1>>4 )&0x0F;
	g_Tiles[dwTile].dwShiftS	= (dwCmd1    )&0x0F;		// LOD stuff

	// Offset by dwTMem - LoadBlock.dwTMem
	g_Tiles[dwTile].dwAddress = g_TI.dwAddr;

	DWORD dwLine = 0;
	switch(g_Tiles[dwTile].dwSize)
	{
		case G_IM_SIZ_4b:  dwLine = (g_Tiles[dwTile].dwLine*8); break;
		case G_IM_SIZ_8b:  dwLine = (g_Tiles[dwTile].dwLine*8); break;
		case G_IM_SIZ_16b: dwLine = (g_Tiles[dwTile].dwLine*8); break;
		case G_IM_SIZ_32b: dwLine = (g_Tiles[dwTile].dwLine*8)*2; break;
	}

	g_Tiles[dwTile].dwPitch = dwLine;

	DL_PF("SetTile: 0x%08x 0x%08x - Tile:%d  Fmt: %s/%s Line:%d (Pitch %d) TMem:0x%04x Palette:%d",
		dwCmd0, dwCmd1,
		dwTile, pszImgFormat[g_Tiles[dwTile].dwFormat], pszImgSize[g_Tiles[dwTile].dwSize],
		g_Tiles[dwTile].dwLine, dwLine, g_Tiles[dwTile].dwTMem, g_Tiles[dwTile].dwPalette);
	DL_PF("         S: Clamp: %s Mirror:%s Mask:0x%x Shift:0x%x",
		pszOnOff[g_Tiles[dwTile].bClampS],pszOnOff[g_Tiles[dwTile].bMirrorS],
		g_Tiles[dwTile].dwMaskS, g_Tiles[dwTile].dwShiftS);
	DL_PF("         T: Clamp: %s Mirror:%s Mask:0x%x Shift:0x%x",
		pszOnOff[g_Tiles[dwTile].bClampT],pszOnOff[g_Tiles[dwTile].bClampT],
		g_Tiles[dwTile].dwMaskT, g_Tiles[dwTile].dwShiftT);

}

void RDP_GFX_SetTileSize(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwTile	= (dwCmd1>>24)&0x07;
		
	DWORD nULS		= (DWORD)(((dwCmd0>>12)&0x0FFF) << 20) >> 20;
	DWORD nULT		= (DWORD)(((dwCmd0    )&0x0FFF) << 20) >> 20;
	DWORD nLRS		= (DWORD)(((dwCmd1>>12)&0x0FFF) << 20) >> 20;
	DWORD nLRT		= (DWORD)(((dwCmd1    )&0x0FFF) << 20) >> 20;

	DL_PF("SetTileSize: 0x%08x 0x%08x Tile:%d (%d,%d) -> (%d,%d) [%d x %d]\n"
		  "             (%d,%d) -> (%d, %d) [%d x %d]",
		dwCmd0, dwCmd1,
		dwTile, nULS/4, nULT/4,
		        nLRS/4, nLRT/4,
				((nLRS/4) - (nULS/4)) + 1,
				((nLRT/4) - (nULT/4)) + 1,
				nULS, nULT, nLRS, nLRT, (nLRS-nULS)+1, (nLRT-nULT)+1);

	// Wetrix hack
	if ((LONG)nULS < 0 || (LONG)nLRS < 0 ||
		(LONG)nULT < 0 || (LONG)nLRT < 0)
	{
		DL_PF("             Ignoring - negative offsets");
		return;
	}
	if (nULT > nLRT || nULS > nLRS)
	{
		DL_PF("             Ignoring - negative size");
		return;
	}

	g_Tiles[dwTile].nULS = nULS / 4;
	g_Tiles[dwTile].nULT = nULT / 4;
	g_Tiles[dwTile].nLRS = nLRS / 4;
	g_Tiles[dwTile].nLRT = nLRT / 4;


	if ((LONG)g_Tiles[dwTile].nULS < 0)
	{
		g_Tiles[dwTile].nLRS -= g_Tiles[dwTile].nULS;
		g_Tiles[dwTile].nULS = 0;
	}
	if ((LONG)g_Tiles[dwTile].nULT < 0)
	{
		g_Tiles[dwTile].nLRT -= g_Tiles[dwTile].nULT;
		g_Tiles[dwTile].nULT = 0;
	}


	g_Tiles[dwTile].nLRS++;
	g_Tiles[dwTile].nLRT++;

	if (g_Tiles[dwTile].nLRS < g_Tiles[dwTile].nULS)
	{
		LONG nTemp = g_Tiles[dwTile].nLRS;
		g_Tiles[dwTile].nLRS = g_Tiles[dwTile].nULS;
		g_Tiles[dwTile].nULS = nTemp;
		
	}

	if (g_Tiles[dwTile].nLRT < g_Tiles[dwTile].nULT)
	{
		LONG nTemp = g_Tiles[dwTile].nLRT;
		g_Tiles[dwTile].nLRT = g_Tiles[dwTile].nULT;
		g_Tiles[dwTile].nULT = nTemp;

	}

	// Testrisphere hack!
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];

	DWORD dwNextCmd0			= *(DWORD *)(g_pu8RamBase + dwPC+0);
	DWORD dwNextCmd1			= *(DWORD *)(g_pu8RamBase + dwPC+4);

	if ((dwNextCmd0>>24) == (BYTE)G_SETTILESIZE &&
		(dwNextCmd1>>24) == dwTile)
	{
		DL_PF("Tetrisphere hack!");
		// Skip it!
		g_dwPCStack[g_dwPCStack.size() - 1] += 8;
	}

	

}


void RDP_GFX_SetTImg(DWORD dwCmd0, DWORD dwCmd1)
{
	static const char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
	static const char *pszImgSize[4] = {"4", "8", "16", "32"};

	g_TI.dwFormat = (dwCmd0>>21)&0x7;
	g_TI.dwSize   = (dwCmd0>>19)&0x3;
	g_TI.dwWidth  = (dwCmd0&0x0FFF) + 1;
	g_TI.dwAddr   = RDPSegAddr(dwCmd1);

	DL_PF("SetTextureImage: Image: 0x%08x Fmt: %s/%s Width: %d", g_TI.dwAddr,
		pszImgFormat[g_TI.dwFormat], pszImgSize[g_TI.dwSize], g_TI.dwWidth);
}


void RDP_GFX_LoadBlock(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwULS		= ((dwCmd0>>12)&0x0FFF)/4;
	DWORD dwULT		= ((dwCmd0    )&0x0FFF)/4;
	DWORD dwPad     = (dwCmd1>>27)&0x1f;
	DWORD dwTile	= (dwCmd1>>24)&0x07;
	DWORD dwPixels	= ((dwCmd1>>12)&0x0FFF) + 1;		// Number of bytes-1?
	DWORD dwDXT		= (dwCmd1    )&0x0FFF;		// 1.11 fixed point

	DWORD dwQWs;
	
	if (dwDXT == 0)
	{
		g_bDXTZero = TRUE;
		dwQWs = 1;
	}
	else
	{
		g_bDXTZero = FALSE;
		dwQWs = 2048 / dwDXT;						// #Quad Words
	}
	DWORD dwBytes = dwQWs * 8;
	DWORD dwWidth = (dwBytes*2)>>g_TI.dwSize;

	DL_PF("LoadBlock: 0x%08x 0x%08x pad(0x%02x)", dwCmd0, dwCmd1, dwPad);
	DL_PF("           Tile:%d (%d,%d) %d pixels DXT:0x%04x = %d QWs => %d pixels/line",
						dwTile, dwULS, dwULT, dwPixels, dwDXT, dwQWs, dwWidth);

	DWORD dwPixelOffset = (dwWidth * dwULT) + dwULS;
	DWORD dwOffset = (dwPixelOffset<<g_TI.dwSize)/2;

	DWORD dwSrcOffset = (g_TI.dwAddr + dwOffset);

	DL_PF("          Offset: 0x%08x", dwSrcOffset);

	// The N64 would load the texture into memory now
	g_nLastLoad = LOAD_BLOCK;
	g_dwLoadAddress = dwSrcOffset;
	g_dwLoadTilePitch = ~0;


}

void RDP_GFX_LoadTile(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwULS		= ((dwCmd0>>12)&0x0FFF)/4;
	DWORD dwULT		= ((dwCmd0    )&0x0FFF)/4;
	DWORD dwTile	=  (dwCmd1>>24)&0x07;
	DWORD dwLRS		= ((dwCmd1>>12)&0x0FFF)/4;
	DWORD dwLRT		= ((dwCmd1    )&0x0FFF)/4;

	// Todo - check if this is odd, and round up?
	DWORD dwPitch = ((g_TI.dwWidth<<g_TI.dwSize)+1)/2;
	DWORD dwOffset = (dwPitch* dwULT) + dwULS;
	DWORD dwSrcOffset = (g_TI.dwAddr + dwOffset);
	
	DL_PF("LoadTile: 0x%08x 0x%08x Tile:%d (%d,%d) -> (%d,%d) [%d x %d]",
		dwCmd0, dwCmd1,
		dwTile, dwULS, dwULT, dwLRS, dwLRT,
		(dwLRS - dwULS)+1, (dwLRT - dwULT)+1);
	DL_PF("          Actual: 0x%08x (LoadTile Pitch: %d)",
		dwSrcOffset, dwPitch);

	g_nLastLoad = LOAD_TILE;
	g_dwLoadAddress = g_TI.dwAddr; 
	g_dwLoadTilePitch = dwPitch;

}

void RDP_GFX_TexRectFlip(DWORD dwCmd0, DWORD dwCmd1)
{ 

	// This command used 128bits, and not 64 bits. This means that we have to look one 
	// Command ahead in the buffer, and update the PC.
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];		// This points to the next instruction
	DWORD dwCmd2 = *(DWORD *)(g_pu8RamBase + dwPC+4);
	DWORD dwCmd3 = *(DWORD *)(g_pu8RamBase + dwPC+4+8);

	// Increment PC so that it points to the right place
	g_dwPCStack[g_dwPCStack.size() - 1] += 16;

	DWORD dwXH		= ((dwCmd0>>12)&0x0FFF)/4;
	DWORD dwYH		= ((dwCmd0    )&0x0FFF)/4;
	DWORD dwTile	= (dwCmd1>>24)&0x07;
	DWORD dwXL		= ((dwCmd1>>12)&0x0FFF)/4;
	DWORD dwYL		= ((dwCmd1    )&0x0FFF)/4;
	DWORD dwS		= (  dwCmd2>>16)&0xFFFF;
	DWORD dwT		= (  dwCmd2    )&0xFFFF;
	LONG  nDSDX 	= (LONG)(SHORT)((  dwCmd3>>16)&0xFFFF);
	LONG  nDTDY	    = (LONG)(SHORT)((  dwCmd3    )&0xFFFF);

	float fS0 = (float)dwS / 32.0f;
	float fT0 = (float)dwT / 32.0f;

	float fDSDX = (float)nDSDX / 1024.0f;
	float fDTDY = (float)nDTDY / 1024.0f;

	if ((g_dwOtherModeH & G_CYC_COPY) == G_CYC_COPY)
	{
		fDSDX /= 4.0f;	// In copy mode 4 pixels are copied at once.
	}
	else
	{
		// Tetrisphere seems to like this:
		if (g_bIncTexRectEdge)
		{
			dwXH++;
			dwYH++;
		}
	}

	float fS1 = fS0 + (fDSDX * (dwYH - dwYL));
	float fT1 = fT0 + (fDTDY * (dwXH - dwXL));
	
	DL_PF("TexRectFlip: 0x%08x 0x%08x Tile:%d (%d,%d) -> (%d,%d)",
		dwCmd0, dwCmd1, dwTile, dwXL, dwYL, dwXH, dwYH);
	DL_PF("         Tex:(%#5f,%#5f) -> (%#5f,%#5f) (DSDX:%#5f DTDY:%#5f)",
		fS0, fT0, fS1, fT1, fDSDX, fDTDY);
	DL_PF("");

	RDP_EnableTexturing(dwTile);
	
	g_Renderer->TexRectFlip(dwXL, dwYL, dwXH, dwYH, fS0, fT0, fS1, fT1);

	g_dwNumTrisRendered += 2;


}


void RDP_GFX_TexRect(DWORD dwCmd0, DWORD dwCmd1)
{
	// This command used 128bits, and not 64 bits. This means that we have to look one 
	// Command ahead in the buffer, and update the PC.
	DWORD dwPC = g_dwPCStack[g_dwPCStack.size() - 1];		// This points to the next instruction
	DWORD dwCmd2 = *(DWORD *)(g_pu8RamBase + dwPC+4);
	DWORD dwCmd3 = *(DWORD *)(g_pu8RamBase + dwPC+4+8);

	// Increment PC so that it points to the right place
	g_dwPCStack[g_dwPCStack.size() - 1] += 16;

	DWORD dwXH		= ((dwCmd0>>12)&0x0FFF)/4;
	DWORD dwYH		= ((dwCmd0    )&0x0FFF)/4;
	DWORD dwTile	= (dwCmd1>>24)&0x07;
	DWORD dwXL		= ((dwCmd1>>12)&0x0FFF)/4;
	DWORD dwYL		= ((dwCmd1    )&0x0FFF)/4;
	DWORD dwS		= (  dwCmd2>>16)&0xFFFF;
	DWORD dwT		= (  dwCmd2    )&0xFFFF;
	LONG  nDSDX 	= (LONG)(SHORT)((  dwCmd3>>16)&0xFFFF);
	LONG  nDTDY	    = (LONG)(SHORT)((  dwCmd3    )&0xFFFF);

	float fS0 = (float)dwS / 32.0f;
	float fT0 = (float)dwT / 32.0f;

	float fDSDX = (float)nDSDX / 1024.0f;
	float fDTDY = (float)nDTDY / 1024.0f;

	if ((g_dwOtherModeH & G_CYC_COPY) == G_CYC_COPY)
	{
		fDSDX /= 4.0f;	// In copy mode 4 pixels are copied at once.
	}
	else
	{
		// Tetrisphere seems to like this:
		if (g_bIncTexRectEdge)
		{
			dwXH++;
			dwYH++;
		}
	}

	// Offset to start of this tile!
	// HACK! For Testrisphere, TexCoords are set to zero by default
	if (dwS != 0)
		fS0 -= (float)g_Tiles[dwTile].nULS;
	if (dwT != 0)
		fT0 -= (float)g_Tiles[dwTile].nULT;

	float fS1 = fS0 + (fDSDX * (dwXH - dwXL));
	float fT1 = fT0 + (fDTDY * (dwYH - dwYL));

	DL_PF("TexRect: 0x%08x 0x%08x Tile:%d (%d,%d) -> (%d,%d)",
		dwCmd0, dwCmd1, dwTile, dwXL, dwYL, dwXH, dwYH);
	DL_PF("         Tex:(%#5f,%#5f) -> (%#5f,%#5f) (DSDX:%#5f DTDY:%#5f)",
		fS0, fT0, fS1, fT1, fDSDX, fDTDY);
	DL_PF("");

	RDP_EnableTexturing(dwTile);

	g_Renderer->TexRect(dwXL, dwYL, dwXH, dwYH, fS0, fT0, fS1, fT1);

	g_dwNumTrisRendered += 2;


}



void RDP_GFX_FillRect(DWORD dwCmd0, DWORD dwCmd1)
{ 
	DWORD x0   = ((dwCmd1>>12)&0xFFF)/4;
	DWORD y0   = ((dwCmd1>>0 )&0xFFF)/4;
	DWORD x1   = ((dwCmd0>>12)&0xFFF)/4;
	DWORD y1   = ((dwCmd0>>0 )&0xFFF)/4;

	// Note, in some modes, the right/bottom lines aren't drawn

	DL_PF("FillRect: (%d,%d) (%d,%d)", x0, y0, x1, y1);

	// TODO - In 1/2cycle mode, skip bottom/right edges!?

	if (g_DI.dwAddr == g_CI.dwAddr)
	{
		// Clear the Z Buffer
		//s_pD3DX->SetClearDepth(depth from set fill color);
		s_pD3DX->Clear(D3DCLEAR_ZBUFFER);

		DL_PF("  Clearing ZBuffer");

	}
	else
	{
		// Clear the screen if large rectangle?
		{

			/*if (((x1+1)-x0) == g_dwViWidth &&
				((y1+1)-y0) == g_dwViHeight &&
				(g_dwFillColor & 0x00FFFFFF) == 0x00000000)
			{
				DL_PF("  Filling Rectangle (screen clear)");
			//	s_pD3DX->Clear(D3DCLEAR_TARGET);

				//DBGConsole_Msg(0, "DL ColorBuffer = 0x%08x", g_CI.dwAddr);
			}
			else*/
			{
				DL_PF("  Filling Rectangle");
				g_Renderer->FillRect(x0, y0, x1+1, y1+1, g_dwFillColor);
			}
		}

	}


}


void RDP_GFX_Texture(DWORD dwCmd0, DWORD dwCmd1)
{
	//DWORD dwBowTie = (dwCmd0>>16)&0xFF;		// Don't care about this
	DWORD dwLevel  = (dwCmd0>>11)&0x07;
	g_dwTextureTile= (dwCmd0>>8 )&0x07;

	BOOL bTextureEnable          = (dwCmd0    )&0x01;
	float fTextureScaleS = (float)((dwCmd1>>16)&0xFFFF) / (65535.0f * 32.0f);
	float fTextureScaleT = (float)((dwCmd1    )&0xFFFF) / (65535.0f * 32.0f);

	g_Renderer->SetTextureEnable(bTextureEnable);
	g_Renderer->SetTextureScale(fTextureScaleS, fTextureScaleT);

	// What happens if these are 0? Interpret as 1.0f?

	DL_PF("Texture: Level: %d Tile: %d %s", dwLevel, g_dwTextureTile, bTextureEnable ? "enabled":"disabled");
	DL_PF("         ScaleS: %f, ScaleT: %f", fTextureScaleS*32.0f, fTextureScaleT*32.0f);

	g_bTextureEnable = bTextureEnable;

}

//802FA128: D7000002 FFFFFFFF CMD Zelda_TEXTURE  texture on=2 tile=0 s=1.00 t=1.00
//000F8AA8: D7000000 FFFFFFFF CMD Zelda_TEXTURE  texture on=0 tile=0 s=1.00 t=1.00

// NOT CHECKED!
void RDP_GFX_ZeldaTexture(DWORD dwCmd0, DWORD dwCmd1)
{
	//DWORD dwBowTie = (dwCmd0>>16)&0xFF;		// Don't care about this
	DWORD dwLevel  = (dwCmd0>>11)&0x07;
	g_dwTextureTile= (dwCmd0>>8 )&0x07;

	BOOL bTextureEnable          = (dwCmd0    )&0x02;		// Important! not 0x01!?!?
	float fTextureScaleS = (float)((dwCmd1>>16)&0xFFFF) / (65535.0f * 32.0f);
	float fTextureScaleT = (float)((dwCmd1    )&0xFFFF) / (65535.0f * 32.0f);

	g_Renderer->SetTextureEnable(bTextureEnable);
	g_Renderer->SetTextureScale(fTextureScaleS, fTextureScaleT);

	// What happens if these are 0? Interpret as 1.0f?

	DL_PF("ZeldaTexture: 0x%08x 0x%08x Level: %d Tile: %d %s", dwCmd0, dwCmd1, dwLevel, g_dwTextureTile, bTextureEnable ? "enabled":"disabled");
	DL_PF("                            ScaleS: %f, ScaleT: %f", fTextureScaleS*32.0f, fTextureScaleT*32.0f);


	g_bTextureEnable = bTextureEnable;

}

void RDP_GFX_SetZImg(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("SetDepthImage: Image: 0x%08x", RDPSegAddr(dwCmd1));

	g_DI.dwAddr = RDPSegAddr(dwCmd1);
}

void RDP_GFX_SetCImg(DWORD dwCmd0, DWORD dwCmd1) {

	static const char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
	static const char *pszImgSize[4] = {"4", "8", "16", "32"};

	DWORD dwFmt   = (dwCmd0>>21)&0x7;
	DWORD dwSiz   = (dwCmd0>>19)&0x3;
	DWORD dwWidth = (dwCmd0&0x0FFF) + 1;

	DL_PF("SetColorImage: Image: 0x%08x", RDPSegAddr(dwCmd1));
	DL_PF("               Fmt: %s Size: %s Width: %d",
		pszImgFormat[dwFmt], pszImgSize[dwSiz], dwWidth);

	g_CI.dwAddr = RDPSegAddr(dwCmd1);
	g_CI.dwFormat = dwFmt;
	g_CI.dwSize = dwSiz;
	g_CI.dwWidth = dwWidth;
}


void RDP_GFX_SetOtherMode_L(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwShift = (dwCmd0>>8)&0xFF;
	DWORD dwLength= (dwCmd0   )&0xFF;
	DWORD dwData  = dwCmd1;

	DWORD dwMask = ((1<<dwLength)-1)<<dwShift;

	// Mask off bits that don't apply. Might not be necessary
	//dwData &= dwMask;

	DWORD dwOldSettings = g_dwOtherModeL;
	
	g_dwOtherModeL = (g_dwOtherModeL&(~dwMask)) | dwData;

	BOOL bZCompare		= (g_dwOtherModeL & Z_CMP)			? TRUE : FALSE;
	BOOL bZUpdate		= (g_dwOtherModeL & Z_UPD)			? TRUE : FALSE;
	
	g_Renderer->SetZCompare( bZCompare );
	g_Renderer->SetZUpdate( bZUpdate );
	

	//FIXME
	BOOL bZModeDecal    = (g_dwOtherModeL & ZMODE_DEC) == ZMODE_DEC     ? TRUE : FALSE;
	g_Renderer->SetZmodeDecal(bZModeDecal);
	/*if (bZModeDecal) {
		g_pD3DDev->SetRenderState( D3DRENDERSTATE_ZBIAS,   1);
	} else {
		g_pD3DDev->SetRenderState( D3DRENDERSTATE_ZBIAS,   0);
	}*/

	DWORD dwAlphaTestMode = (g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE) & 0x3;

	//#define	G_AC_NONE			(0 << G_MDSFT_ALPHACOMPARE)
	//#define	G_AC_THRESHOLD		(1 << G_MDSFT_ALPHACOMPARE)
	//#define	G_AC_DITHER			(3 << G_MDSFT_ALPHACOMPARE)		
	
	if ((dwAlphaTestMode) != 0)
		g_Renderer->SetAlphaTestEnable( TRUE );
	else
		g_Renderer->SetAlphaTestEnable( FALSE );

	




	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		/*if (((g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_ALPHACOMPARE)&0x3))*/
		{
			static const char *alphacomptype[4] = {"None", "Threshold", "?", "Dither"};

			DL_PF("SetOtherMode_L: AlphaCompare: %s", alphacomptype[(g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE)&0x3]);
			//RDP_NOIMPL_WARN("SetOtherMode_L: G_MDSFT_ALPHACOMPARE Not Implemented");

			// Alpha compare
		}

		/*if (((g_dwOtherModeL >> G_MDSFT_ZSRCSEL)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_ZSRCSEL)&0x1))*/
		{
			// Depth source is either pixel or primitive
			DL_PF("SetOtherMode_L: G_MDSFT_ZSRCSEL: %s", ((g_dwOtherModeL >> G_MDSFT_ZSRCSEL)&0x1 ? "Primitive" : "Pixel"));
			//RDP_NOIMPL_WARN("SetOtherMode_L: G_MDSFT_ZSRCSEL Not Implemented");
		}

		/*if (((g_dwOtherModeL >> G_MDSFT_RENDERMODE)&0x1FFFFFFF) !=
			 ((dwOldSettings >> G_MDSFT_RENDERMODE)&0x1FFFFFFF))*/
		{
			BOOL bAA = (g_dwOtherModeL & AA_EN) ? TRUE : FALSE;
			BOOL bZCmp = (g_dwOtherModeL & Z_CMP) ? TRUE : FALSE;
			BOOL bZUpd = (g_dwOtherModeL & Z_UPD) ? TRUE : FALSE;
			BOOL bImRd = (g_dwOtherModeL & IM_RD) ? TRUE : FALSE;
			BOOL bClrOnCvg = (g_dwOtherModeL & CLR_ON_CVG) ? TRUE : FALSE;
			
			BOOL bCvgXAlpha = (g_dwOtherModeL & CVG_X_ALPHA) ? TRUE : FALSE;
			BOOL bAlphaCvgSel = (g_dwOtherModeL & ALPHA_CVG_SEL) ? TRUE : FALSE;
			BOOL bForceBl = (g_dwOtherModeL & FORCE_BL) ? TRUE : FALSE;

			DWORD dwCvgDstMode = (g_dwOtherModeL >> 4) % 0x3;
			DWORD dwZMode = (g_dwOtherModeL >> 6) & 0x3;
			
			static const char * szCvgDstMode[4] = { "Clamp", "Wrap", "Full", "Save" };
			static const char * szZMode[4] = { "Opa", "Inter", "XLU", "Decal" };
			
			// Other G_MDSFT_BLENDER commands here....!

			DL_PF("SetOtherMode_L: G_MDSFT_RENDERMODE\n"
				"AntiAlias:  %s, ZCompare: %s, ZUpdate: %s, ImRd: 0x%02x\n"
				"ClrOnCvg:   0x%02x",
				bAA?"on":"off", bZCmp?"on":"off", bZUpd?"on":"off", bImRd,
				bClrOnCvg);

			DL_PF("CvgDst: %s, ZMode: %s\n"
				  "CvgXAlpha:  0x%02x, AlphaCvgSel: 0x%02x, ForceBL:    0x%02x",
				  szCvgDstMode[dwCvgDstMode], szZMode[dwZMode],
				  bCvgXAlpha, bAlphaCvgSel, bForceBl);
		}

	}
}

void RDP_GFX_ZeldaSetOtherMode_L(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwShift = (dwCmd0>>8)&0xFF;
	DWORD dwLength= (dwCmd0   )&0xFF;
	DWORD dwData  = dwCmd1;

	DWORD dwMask = ((1<<dwLength)-1)<<dwShift;

	// Mask off bits that don't apply. Might not be necessary
	//dwData &= dwMask;

	DWORD dwOldSettings = g_dwOtherModeL;
	
	g_dwOtherModeL = (g_dwOtherModeL&(~dwMask)) | dwData;

	BOOL bZCompare		= (g_dwOtherModeL & Z_CMP)			? TRUE : FALSE;
	BOOL bZUpdate		= (g_dwOtherModeL & Z_UPD)			? TRUE : FALSE;
	
	g_Renderer->SetZCompare( bZCompare );
	g_Renderer->SetZUpdate( bZUpdate );
	

	//FIXME
	BOOL bZModeDecal    = (g_dwOtherModeL & ZMODE_DEC) == ZMODE_DEC     ? TRUE : FALSE;
	g_Renderer->SetZmodeDecal(bZModeDecal);
	/*if (bZModeDecal) {
		g_pD3DDev->SetRenderState( D3DRENDERSTATE_ZBIAS,   1);
	} else {
		g_pD3DDev->SetRenderState( D3DRENDERSTATE_ZBIAS,   0);
	}*/

	DWORD dwAlphaTestMode = (g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE) & 0x3;

	//#define	G_AC_NONE			(0 << G_MDSFT_ALPHACOMPARE)
	//#define	G_AC_THRESHOLD		(1 << G_MDSFT_ALPHACOMPARE)
	//#define	G_AC_DITHER			(3 << G_MDSFT_ALPHACOMPARE)		
	
	if ((dwAlphaTestMode) != 0)
		g_Renderer->SetAlphaTestEnable( TRUE );
	else
		g_Renderer->SetAlphaTestEnable( FALSE );

	




	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		/*if (((g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_ALPHACOMPARE)&0x3))*/
		{
			static const char *alphacomptype[4] = {"None", "Threshold", "?", "Dither"};

			DL_PF("SetOtherMode_L: AlphaCompare: %s", alphacomptype[(g_dwOtherModeL >> G_MDSFT_ALPHACOMPARE)&0x3]);
			//RDP_NOIMPL_WARN("SetOtherMode_L: G_MDSFT_ALPHACOMPARE Not Implemented");

			// Alpha compare
		}

		/*if (((g_dwOtherModeL >> G_MDSFT_ZSRCSEL)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_ZSRCSEL)&0x1))*/
		{
			// Depth source is either pixel or primitive
			DL_PF("SetOtherMode_L: G_MDSFT_ZSRCSEL: %s", ((g_dwOtherModeL >> G_MDSFT_ZSRCSEL)&0x1 ? "Primitive" : "Pixel"));
			//RDP_NOIMPL_WARN("SetOtherMode_L: G_MDSFT_ZSRCSEL Not Implemented");
		}

		/*if (((g_dwOtherModeL >> G_MDSFT_RENDERMODE)&0x1FFFFFFF) !=
			 ((dwOldSettings >> G_MDSFT_RENDERMODE)&0x1FFFFFFF))*/
		{
			BOOL bAA = (g_dwOtherModeL & AA_EN) ? TRUE : FALSE;
			BOOL bZCmp = (g_dwOtherModeL & Z_CMP) ? TRUE : FALSE;
			BOOL bZUpd = (g_dwOtherModeL & Z_UPD) ? TRUE : FALSE;
			BOOL bImRd = (g_dwOtherModeL & IM_RD) ? TRUE : FALSE;
			BOOL bClrOnCvg = (g_dwOtherModeL & CLR_ON_CVG) ? TRUE : FALSE;
			
			BOOL bCvgXAlpha = (g_dwOtherModeL & CVG_X_ALPHA) ? TRUE : FALSE;
			BOOL bAlphaCvgSel = (g_dwOtherModeL & ALPHA_CVG_SEL) ? TRUE : FALSE;
			BOOL bForceBl = (g_dwOtherModeL & FORCE_BL) ? TRUE : FALSE;

			DWORD dwCvgDstMode = (g_dwOtherModeL >> 4) % 0x3;
			DWORD dwZMode = (g_dwOtherModeL >> 6) & 0x3;
			
			static const char * szCvgDstMode[4] = { "Clamp", "Wrap", "Full", "Save" };
			static const char * szZMode[4] = { "Opa", "Inter", "XLU", "Decal" };
			
			// Other G_MDSFT_BLENDER commands here....!

			DL_PF("SetOtherMode_L: G_MDSFT_RENDERMODE\n"
				"AntiAlias:  %s, ZCompare: %s, ZUpdate: %s, ImRd: 0x%02x\n"
				"ClrOnCvg:   0x%02x",
				bAA?"on":"off", bZCmp?"on":"off", bZUpd?"on":"off", bImRd,
				bClrOnCvg);

			DL_PF("CvgDst: %s, ZMode: %s\n"
				  "CvgXAlpha:  0x%02x, AlphaCvgSel: 0x%02x, ForceBL:    0x%02x",
				  szCvgDstMode[dwCvgDstMode], szZMode[dwZMode],
				  bCvgXAlpha, bAlphaCvgSel, bForceBl);
		}

	}
}


void RDP_GFX_SetOtherMode_H(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwShift = (dwCmd0>>8)&0xFF;
	DWORD dwLength= (dwCmd0   )&0xFF;
	DWORD dwData  = dwCmd1;

	DWORD dwMask = ((1<<dwLength)-1)<<dwShift;
	DWORD dwOldSettings = g_dwOtherModeH;

	// Mask off bits that don't apply. Might not be necessary
	dwData &= dwMask;

	g_dwOtherModeH = (g_dwOtherModeH&(~dwMask)) | dwData;


	BOOL bTexPersp    = (g_dwOtherModeH & G_TP_PERSP) ? TRUE : FALSE;
	DWORD dwCycleType = (g_dwOtherModeH>>G_MDSFT_CYCLETYPE)&0x3;
	DWORD dwTextFilt  = (g_dwOtherModeH>>G_MDSFT_TEXTFILT)&0x3;
	
	g_Renderer->SetTexturePerspective( bTexPersp );
	g_Renderer->SetCycleType( dwCycleType );
	

	switch (dwTextFilt<<G_MDSFT_TEXTFILT)
	{
		case G_TF_POINT:
			g_Renderer->SetMinFilter( 0, D3DTFN_POINT );
			g_Renderer->SetMagFilter( 0, D3DTFG_POINT );
			break;
		case G_TF_AVERAGE:	// ?
			g_Renderer->SetMinFilter( 0, D3DTFN_LINEAR );
			g_Renderer->SetMagFilter( 0, D3DTFG_LINEAR );
			break;
		case G_TF_BILERP:
			g_Renderer->SetMinFilter( 0, D3DTFN_LINEAR );
			g_Renderer->SetMagFilter( 0, D3DTFG_LINEAR );
			break;
		default:
			// Unknown!
			g_Renderer->SetMinFilter( 0, D3DTFN_POINT );
			g_Renderer->SetMagFilter( 0, D3DTFG_POINT );
			break;
	}







	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		/*if (((g_dwOtherModeH >> G_MDSFT_BLENDMASK)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_BLENDMASK)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_BLENDMASK -- Not supported by N64");
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_BLENDMASK Not Implemented");
		}*/

		if (((g_dwOtherModeH >> G_MDSFT_ALPHADITHER)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_ALPHADITHER)&0x3))
		{
			static const char *alphadithertypes[4] = {"Pattern", "NotPattern", "Noise", "Disable"};
			DL_PF("SetOtherMode_H: G_MDSFT_ALPHADITHER: %s", alphadithertypes[(g_dwOtherModeH>>G_MDSFT_ALPHADITHER)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_ALPHADITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_RGBDITHER)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_RGBDITHER)&0x3))
		{
			static const char *rgbdithertype[4] = {"MagicSQ", "Bayer", "Noise", "Disable"};
			DL_PF("SetOtherMode_H: G_MDSFT_RGBDITHER: %s", rgbdithertype[(g_dwOtherModeH>>G_MDSFT_RGBDITHER)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_RGBDITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_COMBKEY)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_COMBKEY)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_COMBKEY: %s", ((g_dwOtherModeH>>G_MDSFT_COMBKEY)&0x1 ? "Key" : "None"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_COMBKEY Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTCONV)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTCONV)&0x3))
		{
			static const char *convtype[8] = {"Conv", "?", "?", "?",   "?", "FiltConv", "Filt", "?"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTCONV: %s", convtype[(g_dwOtherModeH>>G_MDSFT_TEXTCONV)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTCONV Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTFILT)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTFILT)&0x3))
		{
			static const char *filtertype[4] = {"Point", "?", "Bilinear", "Average"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTFILT: %s", filtertype[(g_dwOtherModeH>>G_MDSFT_TEXTFILT)&0x3]);
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTLUT)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTLUT)&0x3))
		{
			static const char *textluttype[4] = {"None", "?", "RGBA16", "IA16"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTLUT: %s", textluttype[(g_dwOtherModeH>>G_MDSFT_TEXTLUT)&0x3]);
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTLOD)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_TEXTLOD)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTLOD: %s", ((g_dwOtherModeH>>G_MDSFT_TEXTLOD)&0x1 ? "LOD": "Tile"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTLOD Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTDETAIL)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTDETAIL)&0x3))
		{
			static const char *detailtype[4] = {"Clamp", "Sharpen", "Detail", "?"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTDETAIL: %s", detailtype[(g_dwOtherModeH>>G_MDSFT_TEXTDETAIL)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTDETAIL Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTPERSP)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_TEXTPERSP)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTPERSP: %s", ((g_dwOtherModeH>>G_MDSFT_TEXTPERSP)&0x1 ? "On" : "Off"));
		}
			
		if (((g_dwOtherModeH >> G_MDSFT_CYCLETYPE)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_CYCLETYPE)&0x3))
		{
			static const char *cycletype[4] = {"1Cycle", "2Cycle", "Copy", "Fill"};
			DL_PF("SetOtherMode_H: G_MDSFT_CYCLETYPE: %s", cycletype[(g_dwOtherModeH>>G_MDSFT_CYCLETYPE)&0x3]);
		}

		if (((g_dwOtherModeH >> G_MDSFT_COLORDITHER)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_COLORDITHER)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_COLORDITHER");
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_COLORDITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_PIPELINE)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_PIPELINE)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_PIPELINE: %s", ((g_dwOtherModeH>>G_MDSFT_PIPELINE)&0x1 ? "1Primitive" : "NPrimitive"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_PIPELINE Not Implemented");
		}
	}
}

void RDP_GFX_ZeldaSetOtherMode_H(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwShift = (dwCmd0>>8)&0xFF;
	DWORD dwLength= (dwCmd0   )&0xFF;
	DWORD dwData  = dwCmd1;

	DWORD dwMask = ((1<<dwLength)-1)<<dwShift;
	DWORD dwOldSettings = g_dwOtherModeH;

	// Mask off bits that don't apply. Might not be necessary
	dwData &= dwMask;

	g_dwOtherModeH = (g_dwOtherModeH&(~dwMask)) | dwData;


	BOOL bTexPersp    = (g_dwOtherModeH & G_TP_PERSP) ? TRUE : FALSE;
	DWORD dwCycleType = (g_dwOtherModeH>>G_MDSFT_CYCLETYPE)&0x3;
	DWORD dwTextFilt  = (g_dwOtherModeH>>G_MDSFT_TEXTFILT)&0x3;
	
	g_Renderer->SetTexturePerspective( bTexPersp );
	g_Renderer->SetCycleType( dwCycleType );
	
/*
	switch (dwTextFilt<<G_MDSFT_TEXTFILT)
	{
		case G_TF_POINT:
			g_Renderer->SetMinFilter( 0, D3DTFN_POINT );
			g_Renderer->SetMagFilter( 0, D3DTFG_POINT );
			break;
		case G_TF_AVERAGE:	// ?*/
			g_Renderer->SetMinFilter( 0, D3DTFN_LINEAR );
			g_Renderer->SetMagFilter( 0, D3DTFG_LINEAR );
//			break;
/*		case G_TF_BILERP:
			g_Renderer->SetMinFilter( 0, D3DTFN_LINEAR );
			g_Renderer->SetMagFilter( 0, D3DTFG_LINEAR );
			break;
		default:
			// Unknown!
			g_Renderer->SetMinFilter( 0, D3DTFN_POINT );
			g_Renderer->SetMagFilter( 0, D3DTFG_POINT );
			break;
	}*/


	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		/*if (((g_dwOtherModeH >> G_MDSFT_BLENDMASK)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_BLENDMASK)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_BLENDMASK -- Not supported by N64");
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_BLENDMASK Not Implemented");
		}*/

		if (((g_dwOtherModeH >> G_MDSFT_ALPHADITHER)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_ALPHADITHER)&0x3))
		{
			static const char *alphadithertypes[4] = {"Pattern", "NotPattern", "Noise", "Disable"};
			DL_PF("SetOtherMode_H: G_MDSFT_ALPHADITHER: %s", alphadithertypes[(g_dwOtherModeH>>G_MDSFT_ALPHADITHER)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_ALPHADITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_RGBDITHER)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_RGBDITHER)&0x3))
		{
			static const char *rgbdithertype[4] = {"MagicSQ", "Bayer", "Noise", "Disable"};
			DL_PF("SetOtherMode_H: G_MDSFT_RGBDITHER: %s", rgbdithertype[(g_dwOtherModeH>>G_MDSFT_RGBDITHER)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_RGBDITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_COMBKEY)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_COMBKEY)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_COMBKEY: %s", ((g_dwOtherModeH>>G_MDSFT_COMBKEY)&0x1 ? "Key" : "None"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_COMBKEY Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTCONV)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTCONV)&0x3))
		{
			static const char *convtype[8] = {"Conv", "?", "?", "?",   "?", "FiltConv", "Filt", "?"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTCONV: %s", convtype[(g_dwOtherModeH>>G_MDSFT_TEXTCONV)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTCONV Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTFILT)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTFILT)&0x3))
		{
			static const char *filtertype[4] = {"Point", "?", "Bilinear", "Average"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTFILT: %s", filtertype[(g_dwOtherModeH>>G_MDSFT_TEXTFILT)&0x3]);
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTLUT)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTLUT)&0x3))
		{
			static const char *textluttype[4] = {"None", "?", "RGBA16", "IA16"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTLUT: %s", textluttype[(g_dwOtherModeH>>G_MDSFT_TEXTLUT)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTLUT Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTLOD)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_TEXTLOD)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTLOD: %s", ((g_dwOtherModeH>>G_MDSFT_TEXTLOD)&0x1 ? "LOD": "Tile"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTLOD Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTDETAIL)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_TEXTDETAIL)&0x3))
		{
			static const char *detailtype[4] = {"Clamp", "Sharpen", "Detail", "?"};
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTDETAIL: %s", detailtype[(g_dwOtherModeH>>G_MDSFT_TEXTDETAIL)&0x3]);
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_TEXTDETAIL Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_TEXTPERSP)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_TEXTPERSP)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_TEXTPERSP: %s", ((g_dwOtherModeH>>G_MDSFT_TEXTPERSP)&0x1 ? "On" : "Off"));
		}
			
		if (((g_dwOtherModeH >> G_MDSFT_CYCLETYPE)&0x3) !=
			 ((dwOldSettings >> G_MDSFT_CYCLETYPE)&0x3))
		{
			static const char *cycletype[4] = {"1Cycle", "2Cycle", "Copy", "Fill"};
			DL_PF("SetOtherMode_H: G_MDSFT_CYCLETYPE: %s", cycletype[(g_dwOtherModeH>>G_MDSFT_CYCLETYPE)&0x3]);
		}

		if (((g_dwOtherModeH >> G_MDSFT_COLORDITHER)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_COLORDITHER)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_COLORDITHER");
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_COLORDITHER Not Implemented");
		}

		if (((g_dwOtherModeH >> G_MDSFT_PIPELINE)&0x1) !=
			 ((dwOldSettings >> G_MDSFT_PIPELINE)&0x1))
		{
			DL_PF("SetOtherMode_H: G_MDSFT_PIPELINE: %s", ((g_dwOtherModeH>>G_MDSFT_PIPELINE)&0x1 ? "1Primitive" : "NPrimitive"));
			RDP_NOIMPL_WARN("SetOtherMode_H: G_MDSFT_PIPELINE Not Implemented");
		}
	}
}

static void RDP_GFX_InitGeometryMode()
{
	BOOL bCullFront		= (g_dwGeometryMode & G_CULL_FRONT) ? TRUE : FALSE;
	BOOL bCullBack		= (g_dwGeometryMode & G_CULL_BACK) ? TRUE : FALSE;
	
	BOOL bShade			= (g_dwGeometryMode & G_SHADE) ? TRUE : FALSE;
	BOOL bShadeSmooth	= (g_dwGeometryMode & G_SHADING_SMOOTH) ? TRUE : FALSE;
	
	BOOL bFog			= (g_dwGeometryMode & G_FOG) ? TRUE : FALSE;
	BOOL bTextureGen	= (g_dwGeometryMode & G_TEXTURE_GEN) ? TRUE : FALSE;

	BOOL bLighting      = (g_dwGeometryMode & G_LIGHTING) ? TRUE : FALSE;
	BOOL bZBuffer		= (g_dwGeometryMode & G_ZBUFFER)	? TRUE : FALSE;	

	g_Renderer->SetCullMode(bCullFront, bCullBack);
	
	if (bShade && bShadeSmooth)		g_Renderer->SetShadeMode( D3DSHADE_GOURAUD );
	else							g_Renderer->SetShadeMode( D3DSHADE_FLAT );
	
	g_Renderer->SetFogEnable( bFog );
	g_Renderer->SetTextureGen(bTextureGen);

	g_Renderer->SetLighting( bLighting );
	g_Renderer->ZBufferEnable( bZBuffer );
}

void RDP_GFX_ClearGeometryMode(DWORD dwCmd0, DWORD dwCmd1)
{

	DWORD dwMask = (dwCmd1);
	
	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DL_PF("ClearGeometryMode: Mask=0x%08x", dwMask);
		if (dwMask & G_ZBUFFER)						DL_PF("  Disabling ZBuffer");
		if (dwMask & G_TEXTURE_ENABLE)				DL_PF("  Disabling Texture");
		if (dwMask & G_SHADE)						DL_PF("  Disabling Shade");
		if (dwMask & G_SHADING_SMOOTH)				DL_PF("  Disabling Smooth Shading");
		if (dwMask & G_CULL_FRONT)					DL_PF("  Disabling Front Culling");
		if (dwMask & G_CULL_BACK)					DL_PF("  Disabling Back Culling");
		if (dwMask & G_FOG)							DL_PF("  Disabling Fog");
		if (dwMask & G_LIGHTING)					DL_PF("  Disabling Lighting");
		if (dwMask & G_TEXTURE_GEN)					DL_PF("  Disabling Texture Gen");
		if (dwMask & G_TEXTURE_GEN_LINEAR)			DL_PF("  Disabling Texture Gen Linear");
		if (dwMask & G_LOD)							DL_PF("  Disabling LOD (no impl)");
	}
	g_dwGeometryMode &= ~dwMask;


	//g_bTextureEnable = (g_dwGeometryMode & G_TEXTURE_ENABLE);

	RDP_GFX_InitGeometryMode();
}



void RDP_GFX_SetGeometryMode(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwMask = (dwCmd1);

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DL_PF("SetGeometryMode: Mask=0x%08x", dwMask);
		if (dwMask & G_ZBUFFER)						DL_PF("  Enabling ZBuffer");
		if (dwMask & G_TEXTURE_ENABLE)				DL_PF("  Enabling Texture");
		if (dwMask & G_SHADE)						DL_PF("  Enabling Shade");
		if (dwMask & G_SHADING_SMOOTH)				DL_PF("  Enabling Smooth Shading");
		if (dwMask & G_CULL_FRONT)					DL_PF("  Enabling Front Culling");
		if (dwMask & G_CULL_BACK)					DL_PF("  Enabling Back Culling");
		if (dwMask & G_FOG)							DL_PF("  Enabling Fog");
		if (dwMask & G_LIGHTING)					DL_PF("  Enabling Lighting");
		if (dwMask & G_TEXTURE_GEN)					DL_PF("  Enabling Texture Gen");
		if (dwMask & G_TEXTURE_GEN_LINEAR)			DL_PF("  Enabling Texture Gen Linear");
		if (dwMask & G_LOD)							DL_PF("  Enabling LOD (no impl)");
	}
	g_dwGeometryMode |= dwMask;

	//g_bTextureEnable = (g_dwGeometryMode & G_TEXTURE_ENABLE);

	RDP_GFX_InitGeometryMode();
}

// NOT CHECKED
/*
#define G_ZBUFFER				0x00000001
#define G_TEXTURE_ENABLE		0x00000002	// Microcode use only 
#define G_SHADE					0x00000004	// enable Gouraud interp
#define G_SHADING_SMOOTH		0x00000200	// flat or smooth shaded
#define G_CULL_FRONT			0x00001000
#define G_CULL_BACK				0x00002000
#define G_CULL_BOTH				0x00003000	// To make code cleaner
#define G_FOG					0x00010000
#define G_LIGHTING				0x00020000
#define G_TEXTURE_GEN			0x00040000
#define G_TEXTURE_GEN_LINEAR	0x00080000*/

#define G_ZELDA_ZBUFFER			0x00000001		// Guess
#define G_ZELDA_CULL_BACK		0x00000200
#define G_ZELDA_CULL_FRONT		0x00000400
#define G_ZELDA_LIGHTING		0x00020000
#define G_ZELDA_TEXTURE_GEN		0x00040000
#define G_ZELDA_SHADING_FLAT	0x00080000
//#define G_ZELDA_SHADE			0x00080000

/*
00000000 00101110 00000100 00000101    002E0405  texgen(0) flat(1) cull(2) light(1)
00000000 00100011 00000100 00000101    00230405  texgen(0) flat(0) cull(2) light(1)
           n FFL  cccccccc xxxxxxxL?
00000000 00000000 00000000 00000000
*mode geometry: 002E0405  texgen(0) flat(1) cull(2) light(1)
*mode geometry: 00230405  texgen(0) flat(0) cull(2) light(1)
*mode geometry: 00220405  texgen(0) flat(0) cull(2) light(1)
*mode geometry: 00220405  texgen(0) flat(0) cull(2) light(1)
*mode geometry: 00220405  texgen(0) flat(0) cull(2) light(1)
*mode geometry: 00200404  texgen(0) flat(0) cull(2) light(0)
*mode geometry: 00200204  texgen(0) flat(0) cull(1) light(0)
*mode geometry: 00200000  texgen(0) flat(0) cull(0) light(0)
*mode geometry: 00200000  texgen(0) flat(0) cull(0) light(0)
*mode geometry: 00000000  texgen(0) flat(0) cull(0) light(0)
*/

// Seems to be AND (dwCmd0&0xFFFFFF) OR (dwCmd1&0xFFFFFF)
void RDP_GFX_ZeldaSetGeometryMode(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwAnd = (dwCmd0) & 0x00FFFFFF;
	DWORD dwOr  = (dwCmd1) & 0x00FFFFFF;

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DL_PF("ZeldaSetGeometryMode: 0x%08x 0x%08x =(x & 0x%08x) | 0x%08x", dwCmd0, dwCmd1, dwAnd, dwOr);

		if ((~dwAnd) & G_ZELDA_ZBUFFER)					DL_PF("  Disabling ZBuffer");
	//	if ((~dwAnd) & G_ZELDA_TEXTURE_ENABLE)			DL_PF("  Disabling Texture");
	//	if ((~dwAnd) & G_ZELDA_SHADE)						DL_PF("  Disabling Shade");
		if ((~dwAnd) & G_ZELDA_SHADING_FLAT)			DL_PF("  Disabling Flat Shading");
		if ((~dwAnd) & G_ZELDA_CULL_FRONT)				DL_PF("  Disabling Front Culling");
		if ((~dwAnd) & G_ZELDA_CULL_BACK)					DL_PF("  Disabling Back Culling");
	//	if ((~dwAnd) & G_ZELDA_FOG)						DL_PF("  Disabling Fog");
		if ((~dwAnd) & G_ZELDA_LIGHTING)					DL_PF("  Disabling Lighting");
		if ((~dwAnd) & G_ZELDA_TEXTURE_GEN)				DL_PF("  Disabling Texture Gen");
	//	if ((~dwAnd) & G_ZELDA_TEXTURE_GEN_LINEAR)		DL_PF("  Disabling Texture Gen Linear");
	//	if ((~dwAnd) & G_ZELDA_LOD)						DL_PF("  Disabling LOD (no impl)");

		if (dwOr & G_ZELDA_ZBUFFER)						DL_PF("  Enabling ZBuffer");
	//	if (dwOr & G_ZELDA_TEXTURE_ENABLE)				DL_PF("  Enabling Texture");
	//	if (dwOr & G_ZELDA_SHADE)							DL_PF("  Enabling Shade");
		if (dwOr & G_ZELDA_SHADING_FLAT)				DL_PF("  Enabling Flat Shading");
		if (dwOr & G_ZELDA_CULL_FRONT)					DL_PF("  Enabling Front Culling");
		if (dwOr & G_ZELDA_CULL_BACK)						DL_PF("  Enabling Back Culling");
	//	if (dwOr & G_ZELDA_FOG)							DL_PF("  Enabling Fog");
		if (dwOr & G_ZELDA_LIGHTING)						DL_PF("  Enabling Lighting");
		if (dwOr & G_ZELDA_TEXTURE_GEN)					DL_PF("  Enabling Texture Gen");
	//	if (dwOr & G_ZELDA_TEXTURE_GEN_LINEAR)			DL_PF("  Enabling Texture Gen Linear");
	//	if (dwOr & G_ZELDA_LOD)							DL_PF("  Enabling LOD (no impl)");

	}
	g_dwGeometryMode &= dwAnd;
	g_dwGeometryMode |= dwOr;


	BOOL bCullFront		= (g_dwGeometryMode & G_ZELDA_CULL_FRONT) ? TRUE : FALSE;
	BOOL bCullBack		= (g_dwGeometryMode & G_ZELDA_CULL_BACK) ? TRUE : FALSE;
	
	//BOOL bShade			= (g_dwGeometryMode & G_ZELDA_SHADE) ? TRUE : FALSE;
	BOOL bFlatShade		= (g_dwGeometryMode & G_ZELDA_SHADING_FLAT) ? TRUE : FALSE;
	
	//BOOL bFog			= (g_dwGeometryMode & G_ZELDA_FOG) ? TRUE : FALSE;
	BOOL bTextureGen	= (g_dwGeometryMode & G_ZELDA_TEXTURE_GEN) ? TRUE : FALSE;

	BOOL bLighting      = (g_dwGeometryMode & G_ZELDA_LIGHTING) ? TRUE : FALSE;
	BOOL bZBuffer		= (g_dwGeometryMode & G_ZELDA_ZBUFFER)	? TRUE : FALSE;	

	g_Renderer->SetCullMode(bCullFront, bCullBack);
	
	if (bFlatShade)	g_Renderer->SetShadeMode( D3DSHADE_FLAT );
	else			g_Renderer->SetShadeMode( D3DSHADE_GOURAUD );
	
	//g_Renderer->SetFogEnable( bFog );
	g_Renderer->SetTextureGen(bTextureGen);

	g_Renderer->SetLighting( bLighting );
	g_Renderer->ZBufferEnable( bZBuffer );



	//RDP_GFX_InitGeometryMode();
}



void RDP_GFX_SetCombine(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwMux0 = dwCmd0&0x00FFFFFF;
	DWORD dwMux1 = dwCmd1;

	
	g_Renderer->SetMux(dwMux0, dwMux1);
	

	if (g_hRDPDumpHandle != INVALID_HANDLE_VALUE)
	{
		DWORD aRGB0  = (dwMux0>>20)&0x0F;	// c1 c1		// a0
		DWORD bRGB0  = (dwMux1>>28)&0x0F;	// c1 c2		// b0
		DWORD cRGB0  = (dwMux0>>15)&0x1F;	// c1 c3		// c0
		DWORD dRGB0  = (dwMux1>>15)&0x07;	// c1 c4		// d0

		DWORD aA0    = (dwMux0>>12)&0x07;	// c1 a1		// Aa0
		DWORD bA0    = (dwMux1>>12)&0x07;	// c1 a2		// Ab0
		DWORD cA0    = (dwMux0>>9 )&0x07;	// c1 a3		// Ac0
		DWORD dA0    = (dwMux1>>9 )&0x07;	// c1 a4		// Ad0

		DWORD aRGB1  = (dwMux0>>5 )&0x0F;	// c2 c1		// a1
		DWORD bRGB1  = (dwMux1>>24)&0x0F;	// c2 c2		// b1
		DWORD cRGB1  = (dwMux0    )&0x1F;	// c2 c3		// c1
		DWORD dRGB1  = (dwMux1>>6 )&0x07;	// c2 c4		// d1
		
		DWORD aA1    = (dwMux1>>21)&0x07;	// c2 a1		// Aa1
		DWORD bA1    = (dwMux1>>3 )&0x07;	// c2 a2		// Ab1
		DWORD cA1    = (dwMux1>>18)&0x07;	// c2 a3		// Ac1
		DWORD dA1    = (dwMux1    )&0x07;	// c2 a4		// Ad1
		
		DL_PF("SetCombine: Mux0=0x%08x Mux1=0x%08x", dwMux0, dwMux1);

		DL_PF("            RGB0: (%s - %s) * %s + %s", sc_colcombtypes16[aRGB0], sc_colcombtypes16[bRGB0], sc_colcombtypes32[cRGB0], sc_colcombtypes8[dRGB0]);		
		DL_PF("            A0  : (%s - %s) * %s + %s", sc_colcombtypes8[aA0], sc_colcombtypes8[bA0], sc_colcombtypes8[cA0], sc_colcombtypes8[dA0]);
		DL_PF("            RGB1: (%s - %s) * %s + %s", sc_colcombtypes16[aRGB1], sc_colcombtypes16[bRGB1], sc_colcombtypes32[cRGB1], sc_colcombtypes8[dRGB1]);		
		DL_PF("            A1  : (%s - %s) * %s + %s", sc_colcombtypes8[aA1],  sc_colcombtypes8[bA1], sc_colcombtypes8[cA1],  sc_colcombtypes8[dA1]);

	}
}

void RDP_GFX_SetFillColor(DWORD dwCmd0, DWORD dwCmd1)
{
	// TODO - Check colour image format to work out how this ahould be decoded!
	g_dwFillColor = Convert555ToRGBA((WORD)dwCmd1);

	DL_PF("SetFillColor: Color5551=0x%04x = 0x%08x",
		(WORD)dwCmd1, g_dwFillColor);

}

void RDP_GFX_SetFogColor(DWORD dwCmd0, DWORD dwCmd1)
{
	DL_PF("SetFogColor: Color=0x%08x", dwCmd1);

	DWORD dwFogColor = dwCmd1;

	g_Renderer->SetFogColor( dwFogColor );	
}

void RDP_GFX_SetBlendColor(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwRed		= (dwCmd1>>24)&0xFF;
	DWORD dwGreen	= (dwCmd1>>16)&0xFF;
	DWORD dwBlue	= (dwCmd1>>8)&0xFF;
	DWORD dwAlpha	= (dwCmd1)&0xFF;

	DL_PF("SetBlendColor: 0x%08x 0x%08x R:%d G:%d B:%d A:%d", 
		dwCmd0, dwCmd1, dwRed, dwGreen, dwBlue, dwAlpha);

	g_Renderer->SetAlphaRef(dwAlpha);
}


void RDP_GFX_SetPrimColor(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwM		= (dwCmd0>>8)&0xFF;
	DWORD dwL		= (dwCmd0)&0xFF;
	DWORD dwRed		= (dwCmd1>>24)&0xFF;
	DWORD dwGreen	= (dwCmd1>>16)&0xFF;
	DWORD dwBlue	= (dwCmd1>>8)&0xFF;
	DWORD dwAlpha	= (dwCmd1)&0xFF;

	DL_PF("SetPrimColor: 0x%08x 0x%08x M:%d L:%d R:%d G:%d B:%d A:%d", 
		dwCmd0, dwCmd1, dwM, dwL, dwRed, dwGreen, dwBlue, dwAlpha);

	// Goldeneye moving-circles hack
	// Originally this did mess something else up - not sure what now!
	// Mux0=0x00ffffff Mux1=0xfffdf6fb
//	if (dwAlpha == 0)
//		dwAlpha = 255;

	// Set TFACTOR
	g_Renderer->SetPrimitiveColor( RGBA_MAKE(dwRed, dwGreen, dwBlue, dwAlpha));

	
	// Set material props
	g_Renderer->SetPrimitiveColor( dwRed / 255.0f, dwGreen / 255.0f, dwBlue / 255.0f, dwAlpha / 255.0f);
}

void RDP_GFX_SetEnvColor(DWORD dwCmd0, DWORD dwCmd1)
{
	DWORD dwRed		= (dwCmd1>>24)&0xFF;
	DWORD dwGreen	= (dwCmd1>>16)&0xFF;
	DWORD dwBlue	= (dwCmd1>>8)&0xFF;
	DWORD dwAlpha	= (dwCmd1)&0xFF;
	DL_PF("SetEnvColor: R:%d G:%d B:%d A:%d", dwRed, dwGreen, dwBlue, dwAlpha);

	g_Renderer->SetEnvColor( RGBA_MAKE(dwRed, dwGreen, dwBlue, dwAlpha));
}

